from typing import Any, Callable
from asyncio import Future
from inspect import iscoroutinefunction
from slixmpp.xmlstream import JID

APIHandler = Callable[
    [JID | None, str | None, JID | None, Any],
    Any
]

class APIWrapper(object):
    """Slixmpp API wrapper.

    This class provide a shortened binding to access ``self.api`` from
    plugins without having to specify the plugin name or the global
    :class:`~.APIRegistry`.
    """

    def __init__(self, api: "APIRegistry", name: str) -> None:
        self.api = api
        self.name = name
        if name not in self.api.settings:
            self.api.settings[name] = {}

    @property
    def settings(self) -> dict:
        return self.api.settings[self.name]

    def register(
        self,
        handler: APIHandler | None,
        op: str,
        jid: JID | None = None,
        node: str | None = None,
        default: bool = False,
    ) -> None:
        return self.api.register(handler, self.name, op, jid, node, default)

    def register_default(
        self,
        handler: APIHandler | None,
        op: str,
        jid: JID | None = None,
        node: str | None = None,
    ) -> None:
        return self.api.register_default(handler, self.name, op)

    def run(
        self,
        op: str,
        jid: JID | None = None,
        node: str | None = None,
        ifrom: JID | None = None,
        args: Any = None,
    ) -> Future:
        return self.api.run(self.name, op, jid, node, ifrom, args)

    def restore_default(
        self,
        op: str,
        jid: JID | None = None,
        node: str | None = None,
    ) -> None:
        return self.api.restore_default(self.name, op, jid, node)

    def unregister(
        self,
        op: str,
        jid: JID | None = None,
        node: str | None = None,
    ) -> None:
        return self.api.unregister(self.name, op, jid, node)

    def __getitem__(self, op: str) -> APIHandler:
        def partial(jid=None, node=None, ifrom=None, args=None) -> Future:
            return self.api.run(self.name, op, jid, node, ifrom, args)
        return partial


class APIRegistry(object):
    """API Registry.

    This class is the global Slixmpp API registry, on which any handler will
    be registered.
    """

    def __init__(self, xmpp):
        self._handlers = {}
        self._handler_defaults = {}
        self.xmpp = xmpp
        self.settings = {}

    def _setup(self, ctype: str, op: str):
        """Initialize the API callback dictionaries.

        :param ctype: The name of the API to initialize.
        :param op: The API operation to initialize.
        """
        if ctype not in self.settings:
            self.settings[ctype] = {}
        if ctype not in self._handler_defaults:
            self._handler_defaults[ctype] = {}
        if ctype not in self._handlers:
            self._handlers[ctype] = {}
        if op not in self._handlers[ctype]:
            self._handlers[ctype][op] = {'global': None,
                                         'jid': {},
                                         'node': {}}

    def wrap(self, ctype: str) -> APIWrapper:
        """Return a wrapper object that targets a specific API."""
        return APIWrapper(self, ctype)

    def purge(self, ctype: str) -> None:
        """Remove all information for a given API."""
        del self.settings[ctype]
        del self._handler_defaults[ctype]
        del self._handlers[ctype]

    def run(self, ctype: str, op: str, jid: JID | None = None,
            node: str | None = None, ifrom: JID | None = None,
            args: Any = None) -> Future:
        """Execute an API callback, based on specificity.

        The API callback that is executed is chosen based on the combination
        of the provided JID and node:

        ====== ======= ===================
        JID     node    Handler
        ====== ======= ===================
        Given   Given   Node + JID handler
        Given   None    JID handler
        None    Given   Node handler
        None    None    Global handler
        ====== ======= ===================

        A node handler is responsible for servicing a single node at a single
        JID, while a JID handler may respond for any node at a given JID, and
        the global handler will answer to any JID+node combination.

        Handlers should check that the JID ``ifrom`` is authorized to perform
        the desired action.

        .. versionchanged:: 1.8.0
            ``run()`` always returns a future, if the handler is a coroutine
            the future should be awaited on.

        :param ctype: The name of the API to use.
        :param op: The API operation to perform.
        :param jid: Optionally provide specific JID.
        :param node: Optionally provide specific node.
        :param ifrom: Optionally provide the requesting JID.
        :param args: Optional arguments to the handler.
        """
        self._setup(ctype, op)

        if not jid:
            jid = self.xmpp.boundjid
        elif jid and not isinstance(jid, JID):
            jid = JID(jid)
        elif jid == JID(''):
            jid = self.xmpp.boundjid
        assert jid is not None

        if node is None:
            node = ''

        if self.xmpp.is_component:
            if self.settings[ctype].get('component_bare', False):
                jid_str = jid.bare
            else:
                jid_str = jid.full
        else:
            if self.settings[ctype].get('client_bare', False):
                jid_str = jid.bare
            else:
                jid_str = jid.full

        jid = JID(jid_str)

        handler = self._handlers[ctype][op]['node'].get((jid, node), None)
        if handler is None:
            handler = self._handlers[ctype][op]['jid'].get(jid, None)
        if handler is None:
            handler = self._handlers[ctype][op].get('global', None)

        if handler:
            try:
                if iscoroutinefunction(handler):
                    return self.xmpp.wrap(handler(jid, node, ifrom, args))
                else:
                    future: Future = Future(loop=self.xmpp.loop)
                    result = handler(jid, node, ifrom, args)
                    future.set_result(result)
                    return future
            except TypeError:
                # To preserve backward compatibility, drop the ifrom
                # parameter for existing handlers that don't understand it.
                return handler(jid, node, args)
        future = Future(loop=self.xmpp.loop)
        future.set_result(None)
        return future

    def register(self, handler: APIHandler | None, ctype: str, op: str,
                 jid: JID | None = None, node: str | None = None,
                 default: bool = False):
        """Register an API callback, with JID+node specificity.

        The API callback can later be executed based on the
        specificity of the provided JID+node combination.

        See :meth:`~.APIRegistry.run` for more details.

        :param ctype: The name of the API to use.
        :param op: The API operation to perform.
        :param jid: Optionally provide specific JID.
        :param node: Optionally provide specific node.
        """
        self._setup(ctype, op)
        if jid is None and node is None:
            if handler is None:
                handler = self._handler_defaults[op]
            self._handlers[ctype][op]['global'] = handler
        elif jid is not None and node is None:
            self._handlers[ctype][op]['jid'][jid] = handler
        else:
            self._handlers[ctype][op]['node'][(jid, node)] = handler

        if default:
            self.register_default(handler, ctype, op)

    def register_default(self, handler, ctype: str, op: str):
        """Register a default, global handler for an operation.

        :param handler: The default, global handler for the operation.
        :param ctype: The name of the API to modify.
        :param op: The API operation to use.
        """
        self._setup(ctype, op)
        self._handler_defaults[ctype][op] = handler

    def unregister(self, ctype: str, op: str, jid: JID | None = None,
                   node: str | None = None):
        """Remove an API callback.

        The API callback chosen for removal is based on the
        specificity of the provided JID+node combination.

        See :meth:`~ApiRegistry.run` for more details.

        :param ctype: The name of the API to use.
        :param op: The API operation to perform.
        :param jid: Optionally provide specific JID.
        :param node: Optionally provide specific node.
        """
        self._setup(ctype, op)
        self.register(None, ctype, op, jid, node)

    def restore_default(self, ctype: str, op: str, jid: JID | None = None,
                        node: str | None = None):
        """Reset an API callback to use a default handler.

        :param ctype: The name of the API to use.
        :param op: The API operation to perform.
        :param jid: Optionally provide specific JID.
        :param node: Optionally provide specific node.
        """
        self.unregister(ctype, op, jid, node)
        self.register(self._handler_defaults[ctype][op], ctype, op, jid, node)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.


from slixmpp import Iq, Message, Presence
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.xep_0297 import stanza, Forwarded


class XEP_0297(BasePlugin):

    name = 'xep_0297'
    description = 'XEP-0297: Stanza Forwarding'
    dependencies = {'xep_0030', 'xep_0203'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, Forwarded)

        # While these are marked as iterable, that is just for
        # making it easier to extract the forwarded stanza. There
        # still can be only a single forwarded stanza.
        register_stanza_plugin(Forwarded, Message, iterable=True)
        register_stanza_plugin(Forwarded, Presence, iterable=True)
        register_stanza_plugin(Forwarded, Iq, iterable=True)

        register_stanza_plugin(Forwarded, self.xmpp['xep_0203'].stanza.Delay)

        self.xmpp.register_handler(
            Callback('Forwarded Stanza',
                StanzaPath('message/forwarded'),
                self._handle_forwarded))

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('urn:xmpp:forward:0')

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature='urn:xmpp:forward:0')
        self.xmpp.remove_handler('Forwarded Stanza')

    def forward(self, stanza=None, mto=None, mbody=None, mfrom=None, delay=None):
        stanza.stream = None

        msg = self.xmpp.Message()
        msg['to'] = mto
        msg['from'] = mfrom
        msg['body'] = mbody
        msg['forwarded']['stanza'] = stanza
        if delay is not None:
            msg['forwarded']['delay']['stamp'] = delay
        msg.send()

    def _handle_forwarded(self, msg):
        self.xmpp.event('forwarded_stanza', msg)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from asyncio import Future
from typing import (
    Literal,
)

from slixmpp.jid import JID
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0377 import stanza
from slixmpp.plugins.xep_0191 import BlockItem


log = logging.getLogger(__name__)

BlockedJIDs = (
    JID |
    set[JID] |
    list[JID]
)

class XEP_0377(BasePlugin):
    """XEP-0377: Spam reporting"""

    name = 'xep_0377'
    description = 'XEP-0377: Spam Reporting'
    dependencies = {'xep_0030', 'xep_0191'}
    stanza = stanza

    SPAM = 'urn:xmpp:reporting:spam'
    ABUSE = 'urn:xmpp:reporting:abuse'

    def plugin_init(self):
        register_stanza_plugin(BlockItem, stanza.Report)
        register_stanza_plugin(stanza.Report, stanza.Text)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=stanza.Report.namespace)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(stanza.Report.namespace)

    def report(self,
        jids: BlockedJIDs,
        reason: Literal['spam', 'abuse'] = 'spam',
        ifrom: JID | None = None,
        **iqkwargs,
    ) -> Future:
        """Report a JID or a list of JIDs.

        :param jids: JID(s) to block.
        """
        iq = self.xmpp.make_iq_set(ifrom=ifrom)
        if not isinstance(jids, (set, list)):
            jids = [jids]

        for jid in jids:
            item = BlockItem(parent=iq['block'])
            item['jid'] = jid
            if reason == 'spam':
                item['report']['reason'] = self.SPAM
            else:
                item['report']['reason'] = self.ABUSE

        return iq.send(**iqkwargs)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza import Message
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0071 import XHTML_IM as HTMLIM


register_stanza_plugin(Message, HTMLIM)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Erik Reuterborg Larsson, Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream.stanzabase import ElementBase, ET


class Request(ElementBase):
    namespace = 'urn:xmpp:receipts'
    name = 'request'
    plugin_attrib = 'request_receipt'
    interfaces = {'request_receipt'}
    sub_interfaces = interfaces
    is_extension = True

    def setup(self, xml=None):
        self.xml = ET.Element('')
        return True

    def set_request_receipt(self, val):
        self.del_request_receipt()
        if val:
            parent = self.parent()
            parent._set_sub_text("{%s}request" % self.namespace, keep=True)
            if not parent['id']:
                if parent.stream:
                    parent['id'] = parent.stream.new_id()

    def get_request_receipt(self):
        parent = self.parent()
        if parent.xml.find("{%s}request" % self.namespace) is not None:
            return True
        else:
            return False

    def del_request_receipt(self):
        self.parent()._del_sub("{%s}request" % self.namespace)


class Received(ElementBase):
    namespace = 'urn:xmpp:receipts'
    name = 'received'
    plugin_attrib = 'receipt'
    interfaces = {'receipt'}
    sub_interfaces = interfaces
    is_extension = True

    def setup(self, xml=None):
        self.xml = ET.Element('')
        return True

    def set_receipt(self, value):
        self.del_receipt()
        if value:
            parent = self.parent()
            xml = ET.Element("{%s}received" % self.namespace)
            xml.attrib['id'] = value
            parent.append(xml)

    def get_receipt(self):
        parent = self.parent()
        xml = parent.xml.find("{%s}received" % self.namespace)
        if xml is not None:
            return xml.attrib.get('id', '')
        return ''

    def del_receipt(self):
        self.parent()._del_sub('{%s}received' % self.namespace)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from asyncio import Future
from slixmpp import JID, Iq
from slixmpp.stanza import Presence
from slixmpp.roster import RosterItem
from slixmpp.types import RosterState, RosterDBProtocol, JidStr, ResourceDict

from typing import TYPE_CHECKING, Iterator

if TYPE_CHECKING:
    from slixmpp import BaseXMPP


class RosterNode:

    """
    A roster node is a roster for a single JID.

    Attributes:
        xmpp           -- The main Slixmpp instance.
        jid            -- The JID that owns the roster node.
        db             -- Optional interface to an external datastore.
        auto_authorize -- Determines how authorizations are handled:
                            True  -- Accept all subscriptions.
                            False -- Reject all subscriptions.
                            None  -- Subscriptions must be
                                     manually authorized.
                          Defaults to True.
        auto_subscribe -- Determines if bi-directional subscriptions
                          are created after automatically authorizing
                          a subscription request.
                          Defaults to True
        last_status    -- The last sent presence status that was broadcast
                          to all contact JIDs.

    Methods:
        add           -- Add a JID to the roster.
        update        -- Update a JID's subscription information.
        subscribe     -- Subscribe to a JID.
        unsubscribe   -- Unsubscribe from a JID.
        remove        -- Remove a JID from the roster.
        presence      -- Return presence information for a JID's resources.
        send_presence -- Shortcut for sending a presence stanza.
    """
    xmpp: "BaseXMPP"
    jid: JidStr
    ignore_updates: bool
    auto_authorize: bool
    auto_subscribe: bool
    _jids: dict[str, RosterItem]
    db: RosterDBProtocol | None
    last_status: Presence | None

    def __init__(self, xmpp: "BaseXMPP", jid: JidStr, db: RosterDBProtocol | None = None) -> None:
        """
        Create a roster node for a JID.

        :param xmpp: The main Slixmpp instance.
        :param jid: The JID that owns the roster.
        :param db: Optional interface to an external datastore.
        """
        self.xmpp = xmpp
        self.jid = jid
        self.db = db
        self.ignore_updates = False
        self.auto_authorize = True
        self.auto_subscribe = True
        self.last_status = None
        self._version = ''
        self._jids = {}

        if self.db:
            if hasattr(self.db, 'version'):
                self._version = self.db.version(self.jid)
            for jid in self.db.entries(self.jid):
                self.add(jid)

    @property
    def version(self) -> str:
        """Retrieve the roster's version ID."""
        if self.db and hasattr(self.db, 'version'):
            self._version = self.db.version(self.jid)
        return self._version

    @version.setter
    def version(self, version: str) -> None:
        """Set the roster's version ID."""
        self._version = version
        if self.db and hasattr(self.db, 'set_version'):
            self.db.set_version(self.jid, version)

    def __getitem__(self, key: JidStr) -> RosterItem:
        """
        Return the roster item for a subscribed JID.

        A new item entry will be created if one does not already exist.
        """
        if key is None:
            key = JID('')
        if not isinstance(key, JID):
            key = JID(key)
        key = key.bare
        if key not in self._jids:
            self.add(key, save=True)
        return self._jids[key]

    def __delitem__(self, key: JidStr) -> None:
        """
        Remove a roster item from the local storage.

        To remove an item from the server, use the remove() method.
        """
        if key is None:
            key = JID('')
        if not isinstance(key, JID):
            key = JID(key)
        key = key.bare
        if key in self._jids:
            del self._jids[key]

    def __len__(self) -> int:
        """Return the number of JIDs referenced by the roster."""
        return len(self._jids)

    def keys(self) -> list[str]:
        """Return a list of all subscribed JIDs."""
        return list(self._jids.keys())

    def has_jid(self, jid: JidStr) -> bool:
        """Returns whether the roster has a JID."""
        return jid in self._jids

    def groups(self) -> dict[str, list[str]]:
        """Return a dictionary mapping group names to JIDs."""
        result: dict[str, list[str]] = {}
        for jid in self._jids:
            groups = self._jids[jid]['groups']
            if not groups:
                if '' not in result:
                    result[''] = []
                result[''].append(jid)
            for group in groups:
                if group not in result:
                    result[group] = []
                result[group].append(jid)
        return result

    def __iter__(self) -> Iterator[str]:
        """Iterate over the roster items."""
        return self._jids.__iter__()

    def set_backend(self, db: RosterDBProtocol | None = None, save: bool = True) -> None:
        """
        Set the datastore interface object for the roster node.

        :param db: The new datastore interface.
        :param save: If True, save the existing state to the new
                     backend datastore. Defaults to True.
        """
        self.db = db
        if self.db:
            existing_entries = set(self._jids)
            new_entries = set(self.db.entries(self.jid, {}))

            for jid in existing_entries:
                self._jids[jid].set_backend(db, save)
            for jid in new_entries - existing_entries:
                self.add(jid)

    def add(self, jid: JidStr, name: str = '', groups: list[str] | None = None,
            afrom: bool = False, ato: bool = False, pending_in: bool = False,
            pending_out: bool = False, whitelisted: bool = False,
            save: bool = False) -> None:
        """
        Add a new roster item entry.

        :param jid: The JID for the roster item.
        :param name: An alias for the JID.
        :param groups: A list of group names.
        :param afrom: Indicates if the JID has a subscription state
                      of 'from'. Defaults to False.
        :param ato: Indicates if the JID has a subscription state
                    of 'to'. Defaults to False.
        :param pending_in: Indicates if the JID has sent a subscription
                           request to this connection's JID.
                           Defaults to False.
        :param pending_out: Indicates if a subscription request has been sent
                            to this JID.
                            Defaults to False.
        :param whitelisted: Indicates if a subscription request from this JID
                            should be automatically authorized.
                            Defaults to False.
        :param save: Indicates if the item should be persisted
                     immediately to an external datastore,
                     if one is used.
                     Defaults to False.
        """
        if isinstance(jid, JID):
            key = jid.bare
        else:
            key = jid

        state = RosterState({
            'name': name,
            'groups': groups or [],
            'from': afrom,
            'to': ato,
            'pending_in': pending_in,
            'pending_out': pending_out,
            'whitelisted': whitelisted,
            'subscription': 'none',
            'removed': False,
        })
        self._jids[key] = RosterItem(self.xmpp, jid, self.jid,
                                     state=state, db=self.db,
                                     roster=self)
        if save:
            self._jids[key].save()

    def subscribe(self, jid: JidStr) -> None:
        """
        Subscribe to the given JID.

        :param jid: The JID to subscribe to.
        """
        self[jid].subscribe()

    def unsubscribe(self, jid: JidStr) -> None:
        """
        Unsubscribe from the given JID.

        :param jid: The JID to unsubscribe from.
        """
        self[jid].unsubscribe()

    def remove(self, jid: JidStr) -> Future[Iq] | None:
        """
        Remove a JID from the roster.

        :param jid: The JID to remove.
        """
        self[jid].remove()
        if self.xmpp.is_component:
            return None
        return self.update(jid, subscription='remove')

    def update(self, jid: JidStr, name: str | None = None, subscription=None, groups: list[str] | None = None,
               timeout: int | None = None, callback=None) -> Future[Iq] | None:
        """
        Update a JID's subscription information.

        :param jid: The JID to update.
        :param name: Optional alias for the JID.
        :param subscription: The subscription state. May be one of: 'to',
                             'from', 'both', 'none', or 'remove'.
        :param groups: A list of group names.
        :param timeout: The length of time (in seconds) to wait
                        for a response before continuing if blocking
                        is used. Defaults to self.response_timeout.
        :param callback: Optional reference to a stream handler function.
                         Will be executed when the roster is received.
        """
        if groups is None:
            groups = []

        if name:
            self[jid]['name'] = name
        self[jid]['groups'] = groups
        self[jid].save()

        if self.xmpp.is_component:
            return None

        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['roster']['items'] = {jid: {'name': name,
                                       'subscription': subscription,
                                       'groups': groups}}

        return iq.send(timeout=timeout, callback=callback)

    def presence(self, jid: JID, resource: str | None = None) -> ResourceDict | dict[str, ResourceDict]:
        """
        Retrieve the presence information of a JID.

        May return either all online resources' status, or
        a single resource's status.

        :param jid: The JID to lookup.
        :param resource: Optional resource for returning
                         only the status of a single connection.
        """
        if resource is None:
            return self[jid].resources

        default_presence = ResourceDict({
            'status': '',
            'priority': 0,
            'show': '',
        })
        return self[jid].resources.get(resource,
                                       default_presence)

    def reset(self) -> None:
        """
        Reset the state of the roster to forget any current
        presence information. Useful after a disconnection occurs.
        """
        for jid in self:
            self[jid].reset()

    def send_presence(self, **kwargs) -> None:
        """
        Create, initialize, and send a Presence stanza.

        If no recipient is specified, send the presence immediately.
        Otherwise, forward the send request to the recipient's roster
        entry for processing.

        Arguments:
        :param pshow: The presence's show value.
        :param pstatus: The presence's status message.
        :param ppriority: This connections' priority.
        :param pto: The recipient of a directed presence.
        :param pfrom: The sender of a directed presence, which should
                      be the owner JID plus resource.
        :param ptype: The type of presence, such as 'subscribe'.
        :param pnick: Optional nickname of the presence's sender.
        """
        if self.xmpp.is_component and not kwargs.get('pfrom', ''):
            kwargs['pfrom'] = self.jid
        self.xmpp.send_presence(**kwargs)

    def send_last_presence(self) -> None:
        """Resend a presence with the last status"""
        if self.last_status is None:
            self.send_presence()
        else:
            pres = self.last_status
            if self.xmpp.is_component:
                pres['from'] = self.jid
            else:
                del pres['from']
            pres.send()

    def __repr__(self) -> str:
        return repr(self._jids)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.plugins.base import register_plugin

from . import stanza
from .cam import XEP_0494
from .stanza import ClientType, PermissionEnum, Client

register_plugin(XEP_0494)

__all__ = ["XEP_0494", "Client", "ClientType", "PermissionEnum", "stanza"]



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from asyncio import Future

from slixmpp import JID, Message, Iq
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0425 import stanza


class XEP_0425(BasePlugin):
    '''XEP-0425: Moderated Message Retraction'''

    name = 'xep_0425'
    description = 'XEP-0425: Moderated Message Retraction'
    dependencies = {'xep_0424', 'xep_0421'}
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self) -> None:
        stanza.register_plugins()
        self.xmpp.register_handler(Callback(
            'Moderated Message',
            StanzaPath('message/retract/moderated'),
            self._handle_moderated,
        ))

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(feature=stanza.NS)

    def _handle_moderated(self, message: Message):
        if message['type'] == 'groupchat':
            self.xmpp.event('moderated_message', message)

    def plugin_end(self):
        self.xmpp.plugin['xep_0030'].del_feature(feature=stanza.NS)

    async def moderate(self, room: JID, id: str, reason: str = '', *,
                       ifrom: JID | None = None, **iqkwargs) -> Future[Iq]:
        """
        Moderate a message.

        :param room: Room in which the message to moderate was sent.
        :param id: Stanza-ID of the message to moderate.
        :param reason: An optional reason for the moderation.
        :param ifrom: Which JID the message is to be sent from (for components).
        """
        iq = self.xmpp.make_iq_set(ito=room.bare, ifrom=ifrom)
        iq['moderate']['id'] = id
        iq['moderate']['reason'] = reason
        iq['moderate'].enable('retract')
        await iq.send(**iqkwargs)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2021 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp import JID
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0382 import stanza
from slixmpp.stanza import Message


class XEP_0382(BasePlugin):
    '''XEP-0382: Spoiler Messages'''

    name = 'xep_0382'
    description = 'Spoiler Messages'
    dependencies = {'xep_0030'}
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self) -> None:
        stanza.register_plugins()
        Message.sub_interfaces.add('spoiler')

    def session_bind(self, jid: JID):
        self.xmpp['xep_0030'].add_feature(stanza.NS)

    def plugin_end(self):
        self.xmpp.plugin['xep_0030'].del_feature(feature=stanza.NS)
        Message.sub_interfaces.remove('spoiler')



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp import JID, Message
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream.handler import Callback
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0424 import stanza


DEFAULT_FALLBACK = (
    'This person attempted to retract a previous message, but your client '
    'does not support it.'
)


class XEP_0424(BasePlugin):
    '''XEP-0424: Message Retraction'''

    name = 'xep_0424'
    description = 'XEP-0424: Message Retraction'
    dependencies = {'xep_0422', 'xep_0030', 'xep_0359', 'xep_0428', 'xep_0334'}
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self) -> None:
        stanza.register_plugins()
        self.xmpp.register_handler(Callback(
            "Message Retracted",
            StanzaPath("message/retract"),
            self._handle_retract_message,
        ))

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(feature=stanza.NS)

    def plugin_end(self):
        self.xmpp.plugin['xep_0030'].del_feature(feature=stanza.NS)

    def _handle_retract_message(self, message: Message):
        self.xmpp.event('message_retract', message)

    def send_retraction(self, mto: JID, id: str, mtype: str = 'chat',
                        include_fallback: bool = True,
                        fallback_text: str | None = None, *,
                        mfrom: JID | None = None) -> Message:
        """
        Send a message retraction

        :param JID mto: The JID to retract the message from
        :param str id: Message ID to retract
        :param str mtype: Message type
        :param bool include_fallback: Whether to include a fallback body
        :param str | None fallback_text: The content of the fallback
                                            body. None will set the default value.
        :returns: The sent retraction message.
        """
        if fallback_text is None:
            fallback_text = DEFAULT_FALLBACK
        msg = self.xmpp.make_message(mto=mto, mtype=mtype, mfrom=mfrom)
        if include_fallback:
            msg['body'] = fallback_text
            msg.enable('fallback')
        msg['retract']['id'] = id
        msg.enable('store')
        msg.send()
        return msg



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins import BasePlugin

from . import stanza


class XEP_0511(BasePlugin):
    """
    This plugin add supports for XEP-0511: Link Metadata.

    It does not register any stream handler, it just registers stanza interfaces.
    """

    name = "xep_0511"
    description = "Link Metadata"
    dependencies = set()
    stanza = stanza

    def plugin_init(self) -> None:
        stanza.register_plugin()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import datetime as dt

from slixmpp.xmlstream import ElementBase
from slixmpp.plugins import xep_0082


class EntityTime(ElementBase):

    """
    The <time> element represents the local time for an XMPP agent.
    The time is expressed in UTC to make synchronization easier
    between entities, but the offset for the local timezone is also
    included.

    Example <time> stanzas:
    ::

        <iq type="result">
          <time xmlns="urn:xmpp:time">
            <utc>2011-07-03T11:37:12.234569</utc>
            <tzo>-07:00</tzo>
          </time>
        </iq>

    Stanza Interface:
    ::

        time -- The local time for the entity (updates utc and tzo).
        utc  -- The UTC equivalent to local time.
        tzo  -- The local timezone offset from UTC.
    """

    name = 'time'
    namespace = 'urn:xmpp:time'
    plugin_attrib = 'entity_time'
    interfaces = {'tzo', 'utc', 'time'}
    sub_interfaces = interfaces

    def set_time(self, value: str | dt.datetime) -> None:
        """
        Set both the UTC and TZO fields given a time object.

        :param value: A datetime object or properly formatted
                      string equivalent.
        """
        date = value
        if not isinstance(value, dt.datetime):
            date = xep_0082.parse(value)
        self.set_utc(date)
        self.set_tzo(date.tzinfo)

    def get_time(self) -> dt.datetime:
        """
        Return the entity's local time based on the UTC and TZO data.
        """
        date = self.get_utc()
        tz = self.get_tzo()
        return date.astimezone(tz)

    def del_time(self) -> None:
        """Remove both the UTC and TZO fields."""
        del self['utc']
        del self['tzo']

    def get_tzo(self) -> dt.tzinfo:
        """
        Return the timezone offset from UTC as a tzinfo object.
        """
        tzo = self._get_sub_text('tzo')
        if tzo == '':
            tzo = 'Z'
        time = xep_0082.parse('00:00:00%s' % tzo)
        return time.tzinfo

    def set_tzo(self, value: int | dt.tzinfo) -> None:
        """
        Set the timezone offset from UTC.

        :param value: Either a tzinfo object or the number of
                      seconds (positive or negative) to offset.
        """
        time = xep_0082.time(offset=value)
        if xep_0082.parse(time).tzinfo == dt.timezone.utc:
            self._set_sub_text('tzo', 'Z')
        else:
            self._set_sub_text('tzo', time[-6:])

    def get_utc(self) -> dt.datetime:
        """
        Return the time in UTC as a datetime object.
        """
        value = self._get_sub_text('utc')
        if value == '':
            return xep_0082.parse(xep_0082.datetime())
        return xep_0082.parse('%sZ' % value)

    def set_utc(self, value: str | dt.datetime) -> None:
        """
        Set the time in UTC.

        :param value: A datetime object or properly formatted
                      string equivalent.
        """
        date = value
        if not isinstance(value, dt.datetime):
            date = xep_0082.parse(value)
        date = date.astimezone(dt.timezone.utc)
        value = xep_0082.format_datetime(date)
        self._set_sub_text('utc', value)



"""
Stanza interfaces for XEP-0513: Explicit Mentions

Basic usage:

>>> register_plugin()  # automatically done if you use this plugin
>>> msg = Message()
>>> msg["body"] = "Hello, Alice!"
>>> msg["mention"]["begin"] = 7
>>> msg["mention"]["end"] = 12
>>> msg["mention"]["occupantid"] = "alice@occupant-id"
>>> msg.pretty_print()
<message xmlns="jabber:client">
  <body>Hello, Alice!</body>
  <mention xmlns="urn:xmpp:mentions:0" begin="7" end="12" occupantid="alice@occupant-id" />
</message>
"""

# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import logging
from typing import Literal

from slixmpp.jid import JID, InvalidJID
from slixmpp.stanza.message import Message
from slixmpp.xmlstream import ElementBase, register_stanza_plugin

GroupMentionLiteral = Literal[
    "channel",
    "space",
    "server",
    "moderators",
    "participants",
    "visitors",
    "owner",
    "admin",
    "member",
    "none",
]


NAMESPACE = "urn:xmpp:mentions:0"


class Mention(ElementBase):
    name = "mention"
    namespace = NAMESPACE
    plugin_attrib = "mention"
    plugin_multi_attrib = "mentions"
    interfaces = {
        "begin",
        "end",
        "occupantid",
        "jid",
        "mentions",
        "uri",
    }

    def set_begin(self, begin: int) -> None:
        self._set_attr("begin", str(begin))

    def get_begin(self) -> int | None:
        return _int_or_none(self._get_attr("begin"))

    def set_end(self, end: int) -> None:
        self._set_attr("end", str(end))

    def get_end(self) -> int | None:
        return _int_or_none(self._get_attr("end"))

    def mention_group(self, who: GroupMentionLiteral) -> None:
        """
        Mention a group of participants, see section 3.2 of the XEP.
        """
        self._set_attr("mentions", f"{self.namespace}#{who}")

    def set_jid(self, jid: JID | str) -> None:
        self._set_attr("jid", str(jid))

    def get_jid(self) -> JID | None:
        jid_str = self._get_attr("jid")
        if not jid_str:
            return None
        try:
            return JID(jid_str)
        except InvalidJID:
            log.warning("Failed to parse %s as a JID", jid_str)
            return None


class Active(ElementBase):
    namespace = NAMESPACE
    name = plugin_attrib = "active"


class NoPing(ElementBase):
    namespace = NAMESPACE
    name = plugin_attrib = "noping"


def _int_or_none(v: str | None) -> int | None:
    if not v:
        return None
    try:
        return int(v)
    except ValueError:
        log.warning("Failed to parse %s as an integer", v)
        return None


def register_plugin() -> None:
    register_stanza_plugin(Message, Mention, iterable=True)
    register_stanza_plugin(Mention, NoPing, iterable=True)
    register_stanza_plugin(Mention, Active, iterable=True)


log = logging.getLogger(__name__)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class Invisible(ElementBase):
    name = 'invisible'
    namespace = 'urn:xmpp:invisible:0'
    plugin_attrib = 'invisible'
    interfaces = set()


class Visible(ElementBase):
    name = 'visible'
    namespace = 'urn:xmpp:visible:0'
    plugin_attrib = 'visible'
    interfaces = set()



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.xmlstream import ElementBase


class UserTune(ElementBase):

    name = 'tune'
    namespace = 'http://jabber.org/protocol/tune'
    plugin_attrib = 'tune'
    interfaces = {'artist', 'length', 'rating', 'source',
                  'title', 'track', 'uri'}
    sub_interfaces = interfaces

    def set_length(self, value):
        self._set_sub_text('length', str(value))

    def set_rating(self, value):
        self._set_sub_text('rating', str(value))




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream.handler.callback import Callback
from slixmpp.xmlstream.handler.coroutine_callback import CoroutineCallback
from slixmpp.xmlstream.handler.collector import Collector
from slixmpp.xmlstream.handler.waiter import Waiter
from slixmpp.xmlstream.handler.xmlcallback import XMLCallback
from slixmpp.xmlstream.handler.xmlwaiter import XMLWaiter

__all__ = ['Callback', 'CoroutineCallback', 'Waiter', 'XMLCallback', 'XMLWaiter']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Dann Martens (TOMOTON).
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream.stanzabase import ElementBase
from xml.etree import ElementTree as ET


class RPCQuery(ElementBase):
    name = 'query'
    namespace = 'jabber:iq:rpc'
    plugin_attrib = 'rpc_query'


class MethodCall(ElementBase):
    name = 'methodCall'
    namespace = 'jabber:iq:rpc'
    plugin_attrib = 'method_call'
    interfaces = {'method_name', 'params'}

    def get_method_name(self):
        return self._get_sub_text('methodName')

    def set_method_name(self, value):
        return self._set_sub_text('methodName', value)

    def get_params(self):
        return self.xml.find('{%s}params' % self.namespace)

    def set_params(self, params):
        self.append(params)


class MethodResponse(ElementBase):
    name = 'methodResponse'
    namespace = 'jabber:iq:rpc'
    plugin_attrib = 'method_response'
    interfaces = {'params', 'fault'}

    def get_params(self):
        return self.xml.find('{%s}params' % self.namespace)

    def set_params(self, params):
        self.append(params)

    def get_fault(self) -> ET:
        return self.xml.find('{%s}fault' % self.namespace)

    def set_fault(self, fault: ET) -> None:
        self.append(fault)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0199.stanza import Ping
from slixmpp.plugins.xep_0199.ping import XEP_0199


register_plugin(XEP_0199)

__all__ = ['Ping', 'XEP_0199']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import socket
from queue import Queue


class TestSocket(object):

    """
    A dummy socket that reads and writes to queues instead
    of an actual networking socket.

    Methods:
        next_sent -- Return the next sent stanza.
        recv_data -- Make a stanza available to read next.
        recv      -- Read the next stanza from the socket.
        send      -- Write a stanza to the socket.
        makefile  -- Dummy call, returns self.
        read      -- Read the next stanza from the socket.
    """

    def __init__(self, *args, **kwargs):
        """
        Create a new test socket.

        Arguments:
            Same as arguments for socket.socket
        """
        self.socket = socket.socket(*args, **kwargs)
        self.recv_queue = Queue()
        self.send_queue = Queue()
        self.is_live = False
        self.disconnected = False

    def __getattr__(self, name):
        """
        Return attribute values of internal, dummy socket.

        Some attributes and methods are disabled to prevent the
        socket from connecting to the network.

        Arguments:
            name -- Name of the attribute requested.
        """

        def dummy(*args):
            """Method to do nothing and prevent actual socket connections."""
            return None

        overrides = {'connect': dummy,
                     'close': dummy,
                     'shutdown': dummy}

        return overrides.get(name, getattr(self.socket, name))

    # ------------------------------------------------------------------
    # Testing Interface

    def next_sent(self, timeout=None):
        """
        Get the next stanza that has been 'sent'.

        Arguments:
            timeout -- Optional timeout for waiting for a new value.
        """
        args = {'block': False}
        if timeout is not None:
            args = {'block': True, 'timeout': timeout}
        try:
            return self.send_queue.get(**args)
        except:
            return None

    def recv_data(self, data):
        """
        Add data to the receiving queue.

        Arguments:
            data -- String data to 'write' to the socket to be received
                    by the XMPP client.
        """
        self.recv_queue.put(data)

    def disconnect_error(self):
        """
        Simulate a disconnect error by raising a socket.error exception
        for any current or further socket operations.
        """
        self.disconnected = True

    # ------------------------------------------------------------------
    # Socket Interface

    def recv(self, *args, **kwargs):
        """
        Read a value from the received queue.

        Arguments:
            Placeholders. Same as for socket.Socket.recv.
        """
        if self.disconnected:
            raise socket.error
        return self.read(block=True)

    def send(self, data):
        """
        Send data by placing it in the send queue.

        Arguments:
            data -- String value to write.
        """
        if self.disconnected:
            raise socket.error
        self.send_queue.put(data)
        return len(data)

    # ------------------------------------------------------------------
    # File Socket

    def makefile(self, *args, **kwargs):
        """
        File socket version to use with ElementTree.

        Arguments:
            Placeholders, same as socket.Socket.makefile()
        """
        return self

    def read(self, block=True, timeout=None, **kwargs):
        """
        Implement the file socket interface.

        Arguments:
            block   -- Indicate if the read should block until a
                       value is ready.
            timeout -- Time in seconds a block should last before
                       returning None.
        """
        if self.disconnected:
            raise socket.error
        if timeout is not None:
            block = True
        try:
            return self.recv_queue.get(block, timeout)
        except:
            return None

class TestTransport(object):

    """
    A transport socket that reads and writes to queues instead
    of an actual networking socket.

    Methods:
        next_sent -- Return the next sent stanza.
        recv_data -- Make a stanza available to read next.
        recv      -- Read the next stanza from the socket.
        send      -- Write a stanza to the socket.
        makefile  -- Dummy call, returns self.
        read      -- Read the next stanza from the socket.
    """

    def __init__(self, xmpp):
        self.xmpp = xmpp
        self.socket = TestSocket()
    # ------------------------------------------------------------------
    # Testing Interface

    def next_sent(self, timeout=None):
        """
        Get the next stanza that has been 'sent'.

        Arguments:
            timeout -- Optional timeout for waiting for a new value.
        """
        return self.socket.next_sent()

    def disconnect_error(self):
        """
        Simulate a disconnect error by raising a socket.error exception
        for any current or further socket operations.
        """
        self.socket.disconnect_error()

    # ------------------------------------------------------------------
    # Socket Interface

    def recv(self, *args, **kwargs):
        """
        Read a value from the received queue.

        Arguments:
            Placeholders. Same as for socket.Socket.recv.
        """
        return 

    def write(self, data):
        """
        Send data by placing it in the send queue.

        Arguments:
            data -- String value to write.
        """
        self.socket.send(data)
        return len(data)

    # ------------------------------------------------------------------
    # File Socket

    def makefile(self, *args, **kwargs):
        """
        File socket version to use with ElementTree.

        Arguments:
            Placeholders, same as socket.Socket.makefile()
        """
        return self

    def read(self, block=True, timeout=None, **kwargs):
        """
        Implement the file socket interface.

        Arguments:
            block   -- Indicate if the read should block until a
                       value is ready.
            timeout -- Time in seconds a block should last before
                       returning None.
        """
        return self.socket.recv(block, timeout, **kwargs)

    def get_extra_info(self, name, *args, default=None, **kwargs):
        info = {
            'socket': self.socket
        }
        return info.get(name, default)

    def abort(self, *args, **kwargs):
        return

    def close(self, *args, **kwargs):
        return




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.thirdparty import GPG

from asyncio import Future

from slixmpp.stanza import Presence, Message
from slixmpp.plugins.base import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.xep_0027 import stanza, Signed, Encrypted


def _extract_data(data, kind):
    stripped = []
    begin_headers = False
    begin_data = False
    if isinstance(data, bytes):
        data = data.decode()
    for line in data.split('\n'):
        if not begin_headers and 'BEGIN PGP %s' % kind in line:
            begin_headers = True
            continue
        if begin_headers and line.strip() == '':
            begin_data = True
            continue
        if 'END PGP %s' % kind in line:
            return '\n'.join(stripped)
        if begin_data:
            stripped.append(line)
    return ''


class XEP_0027(BasePlugin):
    """
    XEP-0027: Current Jabber OpenPGP Usage
    """

    name = 'xep_0027'
    description = 'XEP-0027: Current Jabber OpenPGP Usage'
    dependencies = set()
    stanza = stanza
    default_config = {
        'gpg_binary': 'gpg',
        'gpg_home': '',
        'use_agent': True,
        'keyring': None,
        'key_server': 'pgp.mit.edu'
    }

    def plugin_init(self):
        self.gpg = GPG(gnupghome=self.gpg_home,
                       gpgbinary=self.gpg_binary,
                       use_agent=self.use_agent,
                       keyring=self.keyring)

        self.xmpp.add_filter('out', self._sign_presence)

        self._keyids = {}

        self.api.register(self._set_keyid, 'set_keyid', default=True)
        self.api.register(self._get_keyid, 'get_keyid', default=True)
        self.api.register(self._del_keyid, 'del_keyid', default=True)
        self.api.register(self._get_keyids, 'get_keyids', default=True)

        register_stanza_plugin(Presence, Signed)
        register_stanza_plugin(Message, Encrypted)

        self.xmpp.add_event_handler('unverified_signed_presence',
                self._handle_unverified_signed_presence)

        self.xmpp.register_handler(
                Callback('Signed Presence',
                    StanzaPath('presence/signed'),
                    self._handle_signed_presence))

        self.xmpp.register_handler(
                Callback('Encrypted Message',
                    StanzaPath('message/encrypted'),
                    self._handle_encrypted_message))

    def plugin_end(self):
        self.xmpp.remove_handler('Encrypted Message')
        self.xmpp.remove_handler('Signed Presence')
        self.xmpp.del_filter('out', self._sign_presence)
        self.xmpp.del_event_handler('unverified_signed_presence',
                self._handle_unverified_signed_presence)

    def _sign_presence(self, stanza):
        if isinstance(stanza, Presence):
            if stanza['type'] == 'available' or \
                    stanza['type'] in Presence.showtypes:
                stanza['signed'] = stanza['status']
        return stanza

    def sign(self, data, jid=None):
        keyid = self.get_keyid(jid)
        if keyid:
            signed = self.gpg.sign(data, keyid=keyid)
            return _extract_data(signed.data, 'SIGNATURE')

    def encrypt(self, data, jid=None):
        keyid = self.get_keyid(jid)
        if keyid:
            enc = self.gpg.encrypt(data, keyid)
            return _extract_data(enc.data, 'MESSAGE')

    def decrypt(self, data, jid=None):
        template = '-----BEGIN PGP MESSAGE-----\n' + \
                   '\n' + \
                   '%s\n' + \
                   '-----END PGP MESSAGE-----\n'
        dec = self.gpg.decrypt(template % data)
        return dec.data

    def verify(self, data, sig, jid=None):
        template = '-----BEGIN PGP SIGNED MESSAGE-----\n' + \
                   'Hash: SHA1\n' + \
                   '\n' + \
                   '%s\n' + \
                   '-----BEGIN PGP SIGNATURE-----\n' + \
                   '\n' + \
                   '%s\n' + \
                   '-----END PGP SIGNATURE-----\n'
        v = self.gpg.verify(template % (data, sig))
        return v

    def set_keyid(self, jid=None, keyid=None) -> Future:
        """Set a keyid for a specific JID.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.api['set_keyid'](jid, args=keyid)

    def get_keyid(self, jid=None) -> Future:
        """Get a keyid for a jid.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.api['get_keyid'](jid)

    def del_keyid(self, jid=None) -> Future:
        """Delete a keyid.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.api['del_keyid'](jid)

    def get_keyids(self) -> Future:
        """Get stored keyids.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.api['get_keyids']()

    def _handle_signed_presence(self, pres):
        self.xmpp.event('unverified_signed_presence', pres)

    def _handle_unverified_signed_presence(self, pres):
        verified = self.verify(pres['status'], pres['signed'])
        if verified.key_id:
            if not self.get_keyid(pres['from']):
                known_keyids = [e['keyid'] for e in self.gpg.list_keys()]
                if verified.key_id not in known_keyids:
                    self.gpg.recv_keys(self.key_server, verified.key_id)
                self.set_keyid(jid=pres['from'], keyid=verified.key_id)
            self.xmpp.event('signed_presence', pres)

    def _handle_encrypted_message(self, msg):
        self.xmpp.event('encrypted_message', msg)

    # =================================================================

    def _set_keyid(self, jid, node, ifrom, keyid):
        self._keyids[jid] = keyid

    def _get_keyid(self, jid, node, ifrom, keyid):
        return self._keyids.get(jid, None)

    def _del_keyid(self, jid, node, ifrom, keyid):
        if jid in self._keyids:
            del self._keyids[jid]

    def _get_keyids(self, jid, node, ifrom, data):
        return self._keyids




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import socket
import threading
from queue import Queue


class TestLiveSocket(object):

    """
    A live test socket that reads and writes to queues in
    addition to an actual networking socket.

    Methods:
        next_sent -- Return the next sent stanza.
        next_recv -- Return the next received stanza.
        recv_data -- Dummy method to have same interface as TestSocket.
        recv      -- Read the next stanza from the socket.
        send      -- Write a stanza to the socket.
        makefile  -- Dummy call, returns self.
        read      -- Read the next stanza from the socket.
    """

    def __init__(self, *args, **kwargs):
        """
        Create a new, live test socket.

        Arguments:
            Same as arguments for socket.socket
        """
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.recv_buffer = []
        self.recv_queue = Queue()
        self.send_queue = Queue()
        self.send_queue_lock = threading.Lock()
        self.recv_queue_lock = threading.Lock()
        self.is_live = True

    def __getattr__(self, name):
        """
        Return attribute values of internal, live socket.

        Arguments:
            name -- Name of the attribute requested.
        """

        return getattr(self.socket, name)

    # ------------------------------------------------------------------
    # Testing Interface

    def disconnect_error(self):
        """
        Used to simulate a socket disconnection error.

        Not used by live sockets.
        """
        try:
            self.socket.shutdown()
            self.socket.close()
        except:
            pass

    def next_sent(self, timeout=None):
        """
        Get the next stanza that has been sent.

        Arguments:
            timeout -- Optional timeout for waiting for a new value.
        """
        args = {'block': False}
        if timeout is not None:
            args = {'block': True, 'timeout': timeout}
        try:
            return self.send_queue.get(**args)
        except:
            return None

    def next_recv(self, timeout=None):
        """
        Get the next stanza that has been received.

        Arguments:
            timeout -- Optional timeout for waiting for a new value.
        """
        args = {'block': False}
        if timeout is not None:
            args = {'block': True, 'timeout': timeout}
        try:
            if self.recv_buffer:
                return self.recv_buffer.pop(0)
            else:
                return self.recv_queue.get(**args)
        except:
            return None

    def recv_data(self, data):
        """
        Add data to a receive buffer for cases when more than a single stanza
        was received.
        """
        self.recv_buffer.append(data)

    # ------------------------------------------------------------------
    # Socket Interface

    def recv(self, *args, **kwargs):
        """
        Read data from the socket.

        Store a copy in the receive queue.

        Arguments:
            Placeholders. Same as for socket.recv.
        """
        data = self.socket.recv(*args, **kwargs)
        with self.recv_queue_lock:
            self.recv_queue.put(data)
        return data

    def send(self, data):
        """
        Send data on the socket.

        Store a copy in the send queue.

        Arguments:
            data -- String value to write.
        """
        with self.send_queue_lock:
            self.send_queue.put(data)
        return self.socket.send(data)

    # ------------------------------------------------------------------
    # File Socket

    def makefile(self, *args, **kwargs):
        """
        File socket version to use with ElementTree.

        Arguments:
            Placeholders, same as socket.makefile()
        """
        return self

    def read(self, *args, **kwargs):
        """
        Implement the file socket read interface.

        Arguments:
            Placeholders, same as socket.recv()
        """
        return self.recv(*args, **kwargs)

    def clear(self):
        """
        Empty the send queue, typically done once the session has started to
        remove the feature negotiation and log in stanzas.
        """
        with self.send_queue_lock:
            for i in range(0, self.send_queue.qsize()):
                self.send_queue.get(block=False)
        with self.recv_queue_lock:
            for i in range(0, self.recv_queue.qsize()):
                self.recv_queue.get(block=False)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import Iq, Message
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0020 import stanza, FeatureNegotiation
from slixmpp.plugins.xep_0004 import Form


log = logging.getLogger(__name__)


class XEP_0020(BasePlugin):

    name = 'xep_0020'
    description = 'XEP-0020: Feature Negotiation'
    dependencies = {'xep_0004', 'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        self.xmpp['xep_0030'].add_feature(FeatureNegotiation.namespace)

        register_stanza_plugin(FeatureNegotiation, Form)

        register_stanza_plugin(Iq, FeatureNegotiation)
        register_stanza_plugin(Message, FeatureNegotiation)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from asyncio import Future

from slixmpp import JID
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0004 import Form
from slixmpp.plugins.xep_0030 import DiscoInfo
from slixmpp.plugins.xep_0128 import StaticExtendedDisco


class XEP_0128(BasePlugin):

    """
    XEP-0128: Service Discovery Extensions

    Allow the use of data forms to add additional identity
    information to disco#info results.

    Also see <http://www.xmpp.org/extensions/xep-0128.html>.

    :var disco: A reference to the XEP-0030 plugin.
    :var static: Object containing the default set of static
                 node handlers.
    """

    name = 'xep_0128'
    description = 'XEP-0128: Service Discovery Extensions'
    dependencies = {'xep_0030', 'xep_0004'}

    def plugin_init(self):
        """Start the XEP-0128 plugin."""
        self._disco_ops = ['set_extended_info',
                           'add_extended_info',
                           'del_extended_info']

        register_stanza_plugin(DiscoInfo, Form, iterable=True)

        self.disco = self.xmpp['xep_0030']
        self.static = StaticExtendedDisco(self.disco.static)

        self.disco.set_extended_info = self.set_extended_info
        self.disco.add_extended_info = self.add_extended_info
        self.disco.del_extended_info = self.del_extended_info

        for op in self._disco_ops:
            self.api.register(getattr(self.static, op), op, default=True)

    def set_extended_info(self, jid=None, node=None, **kwargs) -> Future:
        """
        Set additional, extended identity information to a node.

        Replaces any existing extended information.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param jid: The JID to modify.
        :param node: The node to modify.
        :param data: Either a form, or a list of forms to use
                     as extended information, replacing any
                     existing extensions.
        """
        return self.api['set_extended_info'](jid, node, None, kwargs)

    def add_extended_info(self, jid=None, node=None, **kwargs) -> Future:
        """
        Add additional, extended identity information to a node.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param jid: The JID to modify.
        :param node: The node to modify.
        :param data: Either a form, or a list of forms to add
                     as extended information.
        """
        return self.api['add_extended_info'](jid, node, None, kwargs)

    def del_extended_info(self, jid: JID | None = None,
                          node: str | None = None, **kwargs) -> Future:
        """
        Remove all extended identity information to a node.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param jid: The JID to modify.
        :param node: The node to modify.
        """
        return self.api['del_extended_info'](jid, node, None, kwargs)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2018 Maxime “pep” Buquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0335 import stanza
from slixmpp.plugins.xep_0335.stanza import JSON_Container
from slixmpp.plugins.xep_0335.json_containers import XEP_0335

register_plugin(XEP_0335)

__all__ = ['stanza', 'JSON_Container', 'XEP_0335']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import asyncio
import logging

from typing import Callable
from slixmpp import JID
from slixmpp.xmlstream import register_stanza_plugin, ElementBase
from slixmpp.plugins.base import BasePlugin, register_plugin
from slixmpp.plugins.xep_0004.stanza import Form


log = logging.getLogger(__name__)


class XEP_0163(BasePlugin):

    """
    XEP-0163: Personal Eventing Protocol (PEP)
    """

    name = 'xep_0163'
    description = 'XEP-0163: Personal Eventing Protocol (PEP)'
    dependencies = {'xep_0030', 'xep_0060', 'xep_0115'}

    def register_pep(self, name, stanza):
        """
        Setup and configure events and stanza registration for
        the given PEP stanza:

            - Add disco feature for the PEP content.
            - Register disco interest in the PEP content.
            - Map events from the PEP content's namespace to the given name.

        :param str name: The event name prefix to use for PEP events.
        :param stanza: The stanza class for the PEP content.
        """
        pubsub_stanza = self.xmpp['xep_0060'].stanza
        register_stanza_plugin(pubsub_stanza.EventItem, stanza)

        self.add_interest(stanza.namespace)
        self.xmpp['xep_0030'].add_feature(stanza.namespace)
        self.xmpp['xep_0060'].map_node_event(stanza.namespace, name)

    def add_interest(self, namespace: str, jid: JID | None = None):
        """
        Mark an interest in a PEP subscription by including a disco
        feature with the '+notify' extension.

        :param namespace: The base namespace to register as an interest, such
                          as 'http://jabber.org/protocol/tune'. This may also
                          be a list of such namespaces.
        :param jid: Optionally specify the JID.
        """
        if not isinstance(namespace, set) and not isinstance(namespace, list):
            namespace = [namespace]

        for ns in namespace:
            self.xmpp['xep_0030'].add_feature('%s+notify' % ns,
                                              jid=jid)
        asyncio.ensure_future(
            self.xmpp['xep_0115'].update_caps(jid, broadcast=False),
            loop=self.xmpp.loop,
        )

    def remove_interest(self, namespace: str, jid: JID | None = None):
        """
        Mark an interest in a PEP subscription by including a disco
        feature with the '+notify' extension.

        :param namespace: The base namespace to remove as an interest, such
                          as 'http://jabber.org/protocol/tune'. This may also
                          be a list of such namespaces.
        :param jid: Optionally specify the JID.
        """
        if not isinstance(namespace, (set, list)):
            namespace = [namespace]

        for ns in namespace:
            self.xmpp['xep_0030'].del_feature(jid=jid,
                                              feature='%s+notify' % namespace)
        asyncio.ensure_future(
            self.xmpp['xep_0115'].update_caps(jid, broadcast=False),
            loop=self.xmpp.loop,
        )

    def publish(self, stanza: ElementBase, node: str | None = None,
                id: str | None = None,
                options: Form | None = None,
                ifrom: JID | None = None,
                callback: Callable | None = None,
                timeout: int | None = None):
        """
        Publish a PEP update.

        This is just a (very) thin wrapper around the XEP-0060 publish()
        method to set the defaults expected by PEP.

        :param stanza: The PEP update stanza to publish.
        :param node: The node to publish the item to. If not specified,
                      the stanza's namespace will be used.
        :param id: Optionally specify the ID of the item.
        :param options:  A form of publish options.
        """
        if node is None:
            node = stanza.namespace
        if id is None:
            id = 'current'

        return self.xmpp['xep_0060'].publish(ifrom, node, id=id,
                                             payload=stanza.xml,
                                             options=options, ifrom=ifrom,
                                             callback=callback,
                                             timeout=timeout)


register_plugin(XEP_0163)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from typing import Awaitable

from slixmpp import Iq, JID
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0257 import stanza


log = logging.getLogger(__name__)


class XEP_0257(BasePlugin):

    name = 'xep_0257'
    description = 'XEP-0257: Client Certificate Management for SASL EXTERNAL'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self) -> None:
        stanza.register_plugins()

    async def get_certs(self, ifrom: JID | None = None,
                        timeout: float | None = None) -> set[tuple[str, str, tuple[str, ...]]]:
        """
        Return the list of cert items.

        :param ifrom: JID to send the stanza from (for components).
        :param timeout: Timeout of the query (in seconds)
        :returns: A list of certificate items as tuple of (name, cert, active users).
        :raises IqTimeout: When the query timeouts
        :raises IqError: If the server encounters an error while processing
                         the query.
        """
        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq['from'] = ifrom
        iq.enable('sasl_certs')
        certs = await iq.send(timeout=timeout)
        return {
            (cert['name'], cert['x509cert'], tuple(cert['users']))
            for cert in certs['sasl_certs']['items']
        }

    def add_cert(self, name: str, cert: str,
                 allow_management: bool = True,
                 ifrom: JID | None = None,
                 timeout: float | None = None) -> Awaitable[Iq]:
        """
        Register a cert with the server.

        :param name: Name of the cert.
        :param cert: Base64 of the cert’s DER.
        :param allow_management: Allow management of this cert by a client
                                 logged in with it.
        :param ifrom: JID to send the stanza from (for components).
        :param timeout: Timeeout of the query (in seconds).
        """
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['from'] = ifrom
        iq['sasl_cert_append']['name'] = name
        iq['sasl_cert_append']['x509cert'] = cert
        iq['sasl_cert_append']['cert_management'] = allow_management
        return iq.send(timeout=timeout)

    def disable_cert(self, name: str, ifrom: JID | None = None,
                     timeout: float | None = None) -> Awaitable[Iq]:
        """
        Disable a cert. Clients using this cert are not immediately disconnected.

        :param name: Name of the cert.
        :param ifrom: JID to send the stanza from (for components).
        :param timeout: Timeeout of the query (in seconds).
        """
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['from'] = ifrom
        iq['sasl_cert_disable']['name'] = name
        return iq.send(timeout=timeout)

    def revoke_cert(self, name: str, ifrom: JID | None = None,
                    timeout: float | None = None) -> Awaitable[Iq]:
        """
        Revoke a cert. Clients using this cert are immediately disconnected.

        :param name: Name of the cert.
        :param ifrom: JID to send the stanza from (for components).
        :param timeout: Timeeout of the query (in seconds).
        """
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['from'] = ifrom
        iq['sasl_cert_revoke']['name'] = name
        return iq.send(timeout=timeout)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
import logging

from slixmpp.stanza import Message
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0482 import stanza


log = logging.getLogger(__name__)


class XEP_0482(BasePlugin):

    """
    XEP-0482: Call Invites

    This plugin defines the stanza elements for Call Invites, as well as new
    events:

    - `call-invite`
    - `call-reject`
    - `call-retract`
    - `call-leave`
    - `call-left`
    """

    name = 'xep_0482'
    description = 'XEP-0482: Call Invites'
    dependencies = set()
    stanza = stanza

    def plugin_init(self):
        stanza.register_plugins()

        for event in ('invite', 'reject', 'retract', 'leave', 'left'):
            self.xmpp.register_handler(
                Callback(f'Call {event}',
                         StanzaPath(f'message/call_{event}'),
                         self._handle_event))

    def _handle_event(self, message: Message):
        for event in ('invite', 'reject', 'retract', 'leave', 'left'):
            if message.get_plugin(f'call_{event}', check=True):
                self.xmpp.event(f'call_{event}')

    def plugin_end(self):
        for event in ('invite', 'reject', 'retract', 'leave', 'left'):
            self.xmpp.remove_handler(f'Call {event}')



from slixmpp.stanza import Message
from slixmpp.xmlstream import ElementBase, register_stanza_plugin
from slixmpp.plugins.xep_0428.stanza import Fallback

NS = "urn:xmpp:reply:0"


class Reply(ElementBase):
    namespace = NS
    name = "reply"
    plugin_attrib = "reply"
    interfaces = {"id", "to"}

    def add_quoted_fallback(self, fallback: str, nickname: str | None = None):
        r"""
        Add plain text fallback for clients not implementing XEP-0461.


        ``msg["reply"].add_quoted_fallback("Some text", "Bob")`` will
        prepend ``> Bob:\n> Some text\n`` to the body of the message, and set the
        fallback_body attributes accordingly, so that clients implementing
        XEP-0461 can hide the fallback text.

        :param fallback: Body of the quoted message.
        :param nickname: Optional, nickname of the quoted participant.
        """
        msg = self.parent()
        quoted = "\n".join("> " + x.strip() for x in fallback.split("\n")) + "\n"
        if nickname:
            quoted = "> " + nickname + ":\n" + quoted
        msg["body"] = quoted + msg["body"]
        fallback_elem = Fallback()
        fallback_elem["for"] = NS
        fallback_elem["body"]["start"] = 0
        fallback_elem["body"]["end"] = len(quoted)
        msg.append(fallback_elem)

    def get_fallback_body(self) -> str:
        """
        Get the string containing the fallback body from the parent.
        """
        msg = self.parent()
        for fallback in msg["fallbacks"]:
            if fallback["for"] == NS:
                break
        else:
            return ""
        start = fallback["body"]["start"]
        end = fallback["body"]["end"]
        body = msg["body"]
        if start <= end:
            return body[start:end]
        else:
            return ""

    def strip_fallback_content(self) -> str:
        """
        Remove the fallback contents from the parent body.
        """
        msg = self.parent()
        for fallback in msg["fallbacks"]:
            if fallback["for"] == NS:
                break
        else:
            return msg["body"]

        start = fallback["body"]["start"]
        end = fallback["body"]["end"]
        body = msg["body"]

        if 0 <= start < end <= len(body):
            return body[:start] + body[end:]
        else:
            return body


def register_plugins():
    register_stanza_plugin(Message, Reply)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import datetime as dt

from slixmpp.xmlstream import ElementBase, register_stanza_plugin
from slixmpp.plugins import xep_0082


class File(ElementBase):
    name = 'file'
    namespace = 'http://jabber.org/protocol/si/profile/file-transfer'
    plugin_attrib = 'file'
    interfaces = {'name', 'size', 'date', 'hash', 'desc'}
    sub_interfaces = {'desc'}

    def set_size(self, value):
        self._set_attr('size', str(value))

    def get_date(self):
        timestamp = self._get_attr('date')
        return xep_0082.parse(timestamp)

    def set_date(self, value):
        if isinstance(value, dt.datetime):
            value = xep_0082.format_datetime(value)
        self._set_attr('date', value)


class Range(ElementBase):
    name = 'range'
    namespace = 'http://jabber.org/protocol/si/profile/file-transfer'
    plugin_attrib = 'range'
    interfaces = {'length', 'offset'}

    def set_length(self, value):
        self._set_attr('length', str(value))

    def set_offset(self, value):
        self._set_attr('offset', str(value))


register_stanza_plugin(File, Range)




# slixmpp: The Slick XMPP Library
# Copyright (C) 2018 Emmanuel Gil Peyrot
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase

class Request(ElementBase):
    plugin_attrib = 'http_upload_request'
    name = 'request'
    namespace = 'urn:xmpp:http:upload:0'
    interfaces = {'filename', 'size', 'content-type'}

class Slot(ElementBase):
    plugin_attrib = 'http_upload_slot'
    name = 'slot'
    namespace = 'urn:xmpp:http:upload:0'

class Put(ElementBase):
    plugin_attrib = 'put'
    name = 'put'
    namespace = 'urn:xmpp:http:upload:0'
    interfaces = {'url'}

class Get(ElementBase):
    plugin_attrib = 'get'
    name = 'get'
    namespace = 'urn:xmpp:http:upload:0'
    interfaces = {'url'}

class Header(ElementBase):
    plugin_attrib = 'header'
    name = 'header'
    namespace = 'urn:xmpp:http:upload:0'
    plugin_multi_attrib = 'headers'
    interfaces = {'name', 'value'}

    def get_value(self):
        return self.xml.text

    def set_value(self, value):
        self.xml.text = value

    def del_value(self):
        self.xml.text = ''



from slixmpp.plugins import BasePlugin

from . import stanza


class XEP_0462(BasePlugin):
    """
    XEP-0462: PubSub Type Filtering

    This plugin makes it possible to filter pubsub items based on their types
    in disco#items queries.

    Registering the plugin sets the disco feature, and registers stanza plugins.
    """

    name = "xep_0462"
    description = "XEP-0462: PubSub Type Filtering"
    dependencies = {"xep_0004", "xep_0030"}
    stanza = stanza

    def plugin_init(self) -> None:
        stanza.register_plugin()
        self.xmpp["xep_0030"].add_feature(stanza.NS)

    def plugin_end(self):
        self.xmpp["xep_0030"].del_feature(stanza.NS)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin
from slixmpp.plugins.xep_0439.stanza import (
    Response, Action, ActionSelected,
)
from slixmpp.plugins.xep_0439.quickresponse import XEP_0439

register_plugin(XEP_0439)

__all__ = ['Response', 'Action', 'ActionSelected', 'XEP_0439']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin
from slixmpp.plugins.xep_0403.stanza import MIXPresence
from slixmpp.plugins.xep_0403.mix_presence import XEP_0403

register_plugin(XEP_0403)

__all__ = ['MIXPresence', 'XEP_0403']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from asyncio import Future

import slixmpp
from slixmpp import JID
from slixmpp.stanza import Iq
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0092 import Version, stanza


log = logging.getLogger(__name__)


class XEP_0092(BasePlugin):

    """
    XEP-0092: Software Version
    """

    name = 'xep_0092'
    description = 'XEP-0092: Software Version'
    dependencies = {'xep_0030'}
    stanza = stanza
    default_config = {
        'software_name': 'Slixmpp',
        'version': slixmpp.__version__,
        'os': ''
    }

    def plugin_init(self):
        """
        Start the XEP-0092 plugin.
        """
        if 'name' in self.config:
            self.software_name = self.config['name']

        self.xmpp.register_handler(
                Callback('Software Version',
                         StanzaPath('iq@type=get/software_version'),
                         self._handle_version))

        register_stanza_plugin(Iq, Version)

    def plugin_end(self):
        self.xmpp.remove_handler('Software Version')
        self.xmpp['xep_0030'].del_feature(feature='jabber:iq:version')

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature('jabber:iq:version')

    def _handle_version(self, iq: Iq):
        """
        Respond to a software version query.

        :param iq: The Iq stanza containing the software version query.
        """
        iq = iq.reply()
        if self.software_name:
            iq['software_version']['name'] = self.software_name
            iq['software_version']['version'] = self.version
            iq['software_version']['os'] = self.os
        else:
            iq.error()
            iq['error']['type'] = 'cancel'
            iq['error']['condition'] = 'service-unavailable'
        iq.send()

    def get_version(self, jid: JID, ifrom: JID | None = None, **iqkwargs) -> Future:
        """
        Retrieve the software version of a remote agent.

        :param jid: The JID of the entity to query.
        """
        iq = self.xmpp.make_iq_get(ito=jid, ifrom=ifrom)
        iq['query'] = Version.namespace
        return iq.send(**iqkwargs)



from slixmpp.plugins.xep_0297 import Forwarded
from slixmpp.stanza import Iq, Message
from slixmpp.xmlstream import ElementBase, register_stanza_plugin

NS = "urn:xmpp:privilege:2"


class Privilege(ElementBase):
    namespace = NS
    name = "privilege"
    plugin_attrib = "privilege"

    def permission(self, access):
        for perm in self["perms"]:
            if perm["access"] == access:
                return perm["type"]

    def roster(self):
        return self.permission("roster")

    def message(self):
        return self.permission("message")

    def presence(self):
        return self.permission("presence")

    def add_perm(self, access, type_):
        # This should only be needed for servers, so maybe out of scope for slixmpp
        perm = Perm()
        perm["type"] = type_
        perm["access"] = access
        self.append(perm)


class Perm(ElementBase):
    namespace = NS
    name = "perm"
    plugin_attrib = "perm"
    plugin_multi_attrib = "perms"
    interfaces = {"type", "access"}


class NameSpace(ElementBase):
    namespace = NS
    name = "namespace"
    plugin_attrib = "namespace"
    plugin_multi_attrib = "namespaces"
    interfaces = {"ns", "type"}


class PrivilegedIq(ElementBase):
    namespace = NS
    name = "privileged_iq"
    plugin_attrib = "privileged_iq"


def register():
    register_stanza_plugin(Message, Privilege)
    register_stanza_plugin(Iq, Privilege)
    register_stanza_plugin(Privilege, Forwarded)
    register_stanza_plugin(Privilege, Perm, iterable=True)
    register_stanza_plugin(Perm, NameSpace, iterable=True)
    register_stanza_plugin(Iq, PrivilegedIq)



from datetime import datetime

from slixmpp.plugins.xep_0082 import format_datetime, parse
from slixmpp.xmlstream import ElementBase

NS = "urn:xmpp:jingle:apps:file-transfer:5"


class File(ElementBase):
    name = "file"
    namespace = NS
    plugin_attrib = "file"
    interfaces = sub_interfaces = {"media-type", "name", "date", "size", "hash", "desc"}

    def set_size(self, size: int):
        self._set_sub_text("size", str(size))

    def get_size(self):
        return _int_or_none(self._get_sub_text("size"))

    def get_date(self):
        try:
            return parse(self._get_sub_text("date"))
        except ValueError:
            return

    def set_date(self, stamp: datetime):
        try:
            self._set_sub_text("date", format_datetime(stamp))
        except ValueError:
            pass


def _int_or_none(v):
    try:
        return int(v)
    except ValueError:
        return None



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import logging

from asyncio import Future
from slixmpp.stanza.message import Message
from slixmpp.stanza.presence import Presence
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0172 import stanza, UserNick


log = logging.getLogger(__name__)


class XEP_0172(BasePlugin):

    """
    XEP-0172: User Nickname
    """

    name = 'xep_0172'
    description = 'XEP-0172: User Nickname'
    dependencies = {'xep_0163'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, UserNick)
        register_stanza_plugin(Presence, UserNick)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=UserNick.namespace)
        self.xmpp['xep_0163'].remove_interest(UserNick.namespace)

    def session_bind(self, jid):
        self.xmpp['xep_0163'].register_pep('user_nick', UserNick)

    def publish_nick(self, nick: str | None = None, **pubsubkwargs) -> Future:
        """
        Publish the user's current nick.

        :param nick: The user nickname to publish.
        """
        nickname = UserNick()
        nickname['nick'] = nick
        return self.xmpp['xep_0163'].publish(
            nickname,
            node=UserNick.namespace,
            **pubsubkwargs
        )

    def stop(self, **pubsubkwargs) -> Future:
        """
        Clear existing user nick information to stop notifications.
        """
        nick = UserNick()
        return self.xmpp['xep_0163'].publish(
            nick,
            node=UserNick.namespace,
            **pubsubkwargs
        )



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import StanzaBase
from typing import ClassVar


class Abort(StanzaBase):

    """
    """

    name = 'abort'
    namespace = 'urn:ietf:params:xml:ns:xmpp-sasl'
    interfaces: ClassVar[set[str]] = set()
    plugin_attrib = name

    def setup(self, xml):
        StanzaBase.setup(self, xml)
        self.xml.tag = self.tag_name()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.stanza import Message
from slixmpp.xmlstream import (
    ElementBase,
    register_stanza_plugin,
)


NS = 'urn:xmpp:sid:0'


class StanzaID(ElementBase):
    namespace = NS
    name = 'stanza-id'
    plugin_attrib = 'stanza_id'
    interfaces = {'id', 'by'}


class OriginID(ElementBase):
    namespace = NS
    name = 'origin-id'
    plugin_attrib = 'origin_id'
    interfaces = {'id'}


def register_plugins():
    register_stanza_plugin(Message, StanzaID)
    register_stanza_plugin(Message, OriginID)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.plugins import BasePlugin, register_plugin


class XEP_0133(BasePlugin):

    name = 'xep_0133'
    description = 'XEP-0133: Service Administration'
    dependencies = {'xep_0030', 'xep_0004', 'xep_0050'}
    commands = {'add-user', 'delete-user', 'disable-user',
                'reenable-user', 'end-user-session', 'get-user-password',
                'change-user-password', 'get-user-roster',
                'get-user-lastlogin', 'user-stats', 'edit-blacklist',
                'edit-whitelist', 'get-registered-users-num',
                'get-disabled-users-num', 'get-online-users-num',
                'get-active-users-num', 'get-idle-users-num',
                'get-registered-users-list', 'get-disabled-users-list',
                'get-online-users-list', 'get-online-users',
                'get-active-users', 'get-idle-userslist', 'announce',
                'set-motd', 'edit-motd', 'delete-motd', 'set-welcome',
                'delete-welcome', 'edit-admin', 'restart', 'shutdown'}

    def get_commands(self, jid=None, **kwargs):
        if jid is None:
            jid = self.xmpp.boundjid.server
        return self.xmpp['xep_0050'].get_commands(jid, **kwargs)


def create_command(name):
    def admin_command(self, jid=None, session=None, ifrom=None):
        if jid is None:
            jid = self.xmpp.boundjid.server
        self.xmpp['xep_0050'].start_command(
                jid=jid,
                node='http://jabber.org/protocol/admin#%s' % name,
                session=session,
                ifrom=ifrom)
    return admin_command


for cmd in XEP_0133.commands:
    setattr(XEP_0133, cmd.replace('-', '_'), create_command(cmd))


register_plugin(XEP_0133)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin
from slixmpp.plugins.xep_0444.reactions import XEP_0444

register_plugin(XEP_0444)

__all__ = ['XEP_0444']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2021 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from collections import abc
import logging
from typing import (
    Iterable,
)

from slixmpp.jid import JID
from slixmpp.xmlstream import ElementBase, ET

log = logging.getLogger(__name__)


class Preferences(ElementBase):
    """MAM Preferences payload.

    .. code-block:: xml

        <iq type='set' id='juliet3'>
          <prefs xmlns='urn:xmpp:mam:2' default='roster'>
            <always>
              <jid>romeo@montague.lit</jid>
            </always>
            <never>
              <jid>montague@montague.lit</jid>
            </never>
          </prefs>
        </iq>

    """
    name = 'prefs'
    namespace = 'urn:xmpp:mam:2'
    plugin_attrib = 'mam_prefs'
    #: Available interfaces:
    #:
    #: - ``default``: Default MAM policy (must be one of 'roster', 'always',
    #:   'never'
    #: - ``always``  (``list[JID]``): list of JIDs to always store
    #:   conversations with.
    #: - ``never``  (``list[JID]``): list of JIDs to never store
    #:   conversations with.
    interfaces = {'default', 'always', 'never'}
    sub_interfaces = {'always', 'never'}

    def get_always(self) -> set[JID]:
        """
        Get a usable set of JIDs the server always stores conversations for.
        """
        results = set()

        jids = self.xml.findall('{%s}always/{%s}jid' % (
            self.namespace, self.namespace))

        for jid in jids:
            results.add(JID(jid.text))

        return results

    def set_always(self, value: Iterable[JID]):
        """
        Set the MAM policy to always archive for the specified JIDs.

        :param value: List of JIDs.
        """
        if not isinstance(value, abc.Iterable):
            if isinstance(value, JID):
                log.warning("Wrong type provided to set_always(): %s", value)
                value = [value]
            else:
                raise ValueError(f'Wrong type for "value" ({value}).')
        elif isinstance(value, str):
            raise ValueError('Wrong type for "value" (str).')

        self._set_sub_text('always', '', keep=True)
        always = self.xml.find('{%s}always' % self.namespace)
        if always is None:
            raise ValueError("<always/> was not found, it should not happen")
        always.clear()

        for jid in value:
            jid_xml = ET.Element('{%s}jid' % self.namespace)
            jid_xml.text = str(jid)
            always.append(jid_xml)

    def get_never(self) -> set[JID]:
        """
        Get a usable set of JIDs the server never stores conversations for.
        """
        results = set()

        jids = self.xml.findall('{%s}never/{%s}jid' % (
            self.namespace, self.namespace))

        for jid in jids:
            results.add(JID(jid.text))

        return results

    def set_never(self, value: Iterable[JID]):
        """
        Set the MAM policy to never archive for the specified JIDs.

        :param value: List of JIDs.
        """
        if not isinstance(value, abc.Iterable):
            if isinstance(value, JID):
                log.warning("Wrong type provided to set_never(): %s", value)
                value = [value]
            else:
                raise ValueError(f'Wrong type for "value" ({value}).')
        elif isinstance(value, str):
            raise ValueError('Wrong type for "value" (str).')

        self._set_sub_text('never', '', keep=True)
        never = self.xml.find('{%s}never' % self.namespace)
        never.clear()
        for jid in value:
            jid_xml = ET.Element('{%s}jid' % self.namespace)
            jid_xml.text = str(jid)
            never.append(jid_xml)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import JID
from slixmpp.stanza import Iq
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import CoroutineCallback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0054 import VCardTemp, stanza


log = logging.getLogger(__name__)


class XEP_0054(BasePlugin):

    """
    XEP-0054: vcard-temp
    """

    name = 'xep_0054'
    description = 'XEP-0054: vcard-temp'
    dependencies = {'xep_0030', 'xep_0082'}
    stanza = stanza

    def plugin_init(self):
        """
        Start the XEP-0054 plugin.
        """
        register_stanza_plugin(Iq, VCardTemp)


        self.api.register(self._set_vcard, 'set_vcard', default=True)
        self.api.register(self._get_vcard, 'get_vcard', default=True)
        self.api.register(self._del_vcard, 'del_vcard', default=True)

        self._vcard_cache = {}

        self.xmpp.register_handler(
                CoroutineCallback('VCardTemp',
                    StanzaPath('iq/vcard_temp'),
                    self._handle_get_vcard))

    def plugin_end(self):
        self.xmpp.remove_handler('VCardTemp')
        self.xmpp['xep_0030'].del_feature(feature='vcard-temp')

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('vcard-temp')

    def make_vcard(self) -> VCardTemp:
        """Return an empty vcard element."""
        return VCardTemp()

    async def get_vcard(self, jid: JID | None = None, *,
                        local: bool | None = None, cached: bool = False,
                        ifrom: JID | None = None,
                        **iqkwargs) -> Iq:
        """Retrieve a VCard.

        .. versionchanged:: 1.8.0
            This function is now a coroutine.

        :param jid: JID of the entity to fetch the VCard from.
        :param local: Only check internally for a vcard.
        :param cached: Whether to check in the local cache before
                       sending a query.
        """
        if local is None:
            if jid is not None and not isinstance(jid, JID):
                jid = JID(jid)
                if self.xmpp.is_component:
                    if jid.domain == self.xmpp.boundjid.domain:
                        local = True
                else:
                    if str(jid) == str(self.xmpp.boundjid):
                        local = True
                jid = jid.full
            elif jid in (None, ''):
                local = True

        if local:
            vcard = await self.api['get_vcard'](jid, None, ifrom)
            if not isinstance(vcard, Iq):
                iq = self.xmpp.Iq()
                if vcard is None:
                    vcard = VCardTemp()
                iq.append(vcard)
                return iq
            return vcard

        if cached:
            vcard = await self.api['get_vcard'](jid, None, ifrom)
            if vcard is not None:
                if not isinstance(vcard, Iq):
                    iq = self.xmpp.Iq()
                    iq.append(vcard)
                    return iq
                return vcard

        iq = self.xmpp.make_iq_get(ito=jid, ifrom=ifrom)
        iq.enable('vcard_temp')
        return await iq.send(**iqkwargs)

    async def publish_vcard(self, vcard: VCardTemp | None = None,
                            jid: JID | None = None,
                            ifrom: JID | None = None, **iqkwargs):
        """Publish a vcard.

        .. versionchanged:: 1.8.0
            This function is now a coroutine.

        :param vcard: The VCard to publish.
        :param jid: The JID to publish the VCard to.
        """
        await self.api['set_vcard'](jid, None, ifrom, vcard)
        if self.xmpp.is_component:
            return

        iq = self.xmpp.make_iq_set(ito=jid, ifrom=ifrom)
        iq.append(vcard)
        await iq.send(**iqkwargs)

    async def _handle_get_vcard(self, iq: Iq):
        if iq['type'] == 'result':
            await self.api['set_vcard'](jid=iq['from'], args=iq['vcard_temp'])
            return
        elif iq['type'] == 'get' and self.xmpp.is_component:
            vcard = await self.api['get_vcard'](iq['to'].bare, ifrom=iq['from'])
            if vcard is None:
                raise XMPPError("item-not-found")
            elif isinstance(vcard, Iq):
                await vcard.send()
            else:
                iq = iq.reply()
                iq.append(vcard)
                iq.send()
        elif iq['type'] == 'set':
            raise XMPPError('service-unavailable')

    # =================================================================

    def _set_vcard(self, jid, node, ifrom, vcard):
        self._vcard_cache[jid.bare] = vcard

    def _get_vcard(self, jid, node, ifrom, vcard):
        return self._vcard_cache.get(jid.bare, None)

    def _del_vcard(self, jid, node, ifrom, vcard):
        if jid.bare in self._vcard_cache:
            del self._vcard_cache[jid.bare]



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import PluginManager, PluginNotFound, BasePlugin
from slixmpp.plugins.base import register_plugin, load_plugin


PLUGINS = [
    # XEPS
    'xep_0004',  # Data Forms
    'xep_0009',  # Jabber-RPC
    'xep_0012',  # Last Activity
    'xep_0013',  # Flexible Offline Message Retrieval
#   'xep_0016',  # Privacy Lists. Don’t automatically load
    'xep_0020',  # Feature Negotiation
    'xep_0027',  # Current Jabber OpenPGP Usage
    'xep_0030',  # Service Discovery
    'xep_0033',  # Extended Stanza Addresses
    'xep_0045',  # Multi-User Chat (Client)
    'xep_0047',  # In-Band Bytestreams
#   'xep_0048',  # Legacy Bookmarks. Don’t automatically load
    'xep_0049',  # Private XML Storage
    'xep_0050',  # Ad-hoc Commands
    'xep_0054',  # vcard-temp
    'xep_0055',  # Jabber Search
    'xep_0059',  # Result Set Management
    'xep_0060',  # Pubsub (Client)
    'xep_0065',  # SOCKS5 Bytestreams
    'xep_0066',  # Out of Band Data
    'xep_0070',  # Verifying HTTP Requests via XMPP
    'xep_0071',  # XHTML-IM
    'xep_0077',  # In-Band Registration
#   'xep_0078',  # Non-SASL auth. Don’t automatically load
    'xep_0079',  # Advanced Message Processing
    'xep_0080',  # User Location
    'xep_0082',  # XMPP Date and Time Profiles
    'xep_0084',  # User Avatar
    'xep_0085',  # Chat State Notifications
    'xep_0086',  # Legacy Error Codes
#   'xep_0091',  # Legacy Delayed Delivery. Don’t automatically load
    'xep_0092',  # Software Version
#   'xep_0095',  # Legacy Stream Initiation. Don’t automatically load
#   'xep_0096',  # Legacy SI File Transfer. Don’t automatically load
    'xep_0100',  # Gateway interaction
    'xep_0106',  # JID Escaping
    'xep_0107',  # User Mood
    'xep_0108',  # User Activity
    'xep_0115',  # Entity Capabilities
    'xep_0118',  # User Tune
    'xep_0122',  # Data Forms Validation
    'xep_0128',  # Extended Service Discovery
    'xep_0131',  # Standard Headers and Internet Metadata
    'xep_0133',  # Service Administration
    'xep_0152',  # Reachability Addresses
    'xep_0153',  # vCard-Based Avatars
    'xep_0163',  # Personal Eventing Protocol
    'xep_0172',  # User Nickname
    'xep_0184',  # Message Receipts
    'xep_0186',  # Invisible Command
    'xep_0191',  # Blocking Command
    'xep_0196',  # User Gaming
    'xep_0198',  # Stream Management
    'xep_0199',  # Ping
    'xep_0202',  # Entity Time
    'xep_0203',  # Delayed Delivery
    'xep_0221',  # Data Forms Media Element
    'xep_0222',  # Persistent Storage of Public Data via Pubsub
    'xep_0223',  # Persistent Storage of Private Data via Pubsub
    'xep_0224',  # Attention
    'xep_0231',  # Bits of Binary
    'xep_0235',  # OAuth Over XMPP
#   'xep_0242',  # XMPP Client Compliance 2009. Don’t automatically load
    'xep_0249',  # Direct MUC Invitations
    'xep_0256',  # Last Activity in Presence
    'xep_0257',  # Client Certificate Management for SASL EXTERNAL
    'xep_0258',  # Security Labels in XMPP
    'xep_0264',  # Jingle Content Thumbnails
#   'xep_0270',  # XMPP Compliance Suites 2010. Don’t automatically load
    'xep_0279',  # Server IP Check
    'xep_0280',  # Message Carbons
    'xep_0292',  # vCard4 Over XMPP
    'xep_0297',  # Stanza Forwarding
    'xep_0300',  # Use of Cryptographic Hash Functions in XMPP
#   'xep_0302',  # XMPP Compliance Suites 2012. Don’t automatically load
    'xep_0308',  # Last Message Correction
    'xep_0313',  # Message Archive Management
    'xep_0317',  # Hats
    'xep_0319',  # Last User Interaction in Presence
#   'xep_0323',  # IoT Systems Sensor Data. Don’t automatically load
#   'xep_0325',  # IoT Systems Control. Don’t automatically load
    'xep_0332',  # HTTP Over XMPP Transport
    'xep_0333',  # Chat Markers
    'xep_0334',  # Message Processing Hints
    'xep_0335',  # JSON Containers
    'xep_0352',  # Client State Indication
    'xep_0353',  # Jingle Message Initiation
    'xep_0356',  # Privileged entity
    'xep_0359',  # Unique and Stable Stanza IDs
    'xep_0363',  # HTTP File Upload
    'xep_0369',  # MIX-CORE
    'xep_0377',  # Spam reporting
    'xep_0380',  # Explicit Message Encryption
    'xep_0382',  # Spoiler Messages
    'xep_0385',  # Stateless Inline Media Sharing (SIMS)
    'xep_0394',  # Message Markup
    'xep_0402',  # PEP Native Bookmarks
    'xep_0403',  # MIX-Presence
    'xep_0404',  # MIX-Anon
    'xep_0405',  # MIX-PAM
    'xep_0410',  # MUC Self-ping
    'xep_0421',  # Anonymous unique occupant identifiers for MUCs
    'xep_0422',  # Message Fastening
    'xep_0424',  # Message Retraction
    'xep_0425',  # Moderated Message Retraction
    'xep_0428',  # Message Fallback
    'xep_0437',  # Room Activity Indicators
    'xep_0439',  # Quick Response
    'xep_0441',  # Message Archive Management Preferences
    'xep_0444',  # Message Reactions
    'xep_0446',  # File metadata element
    'xep_0447',  # Stateless file sharing
    'xep_0449',  # Stickers
    'xep_0455',  # Service Outage Status
    'xep_0461',  # Message Replies
    'xep_0462',  # PubSub Type Filtering
    'xep_0463',  # MUC Affiliation Versioning
    'xep_0469',  # Bookmarks Pinning
    'xep_0482',  # Call Invites
    'xep_0490',  # Message Displayed Synchronization
    'xep_0492',  # Chat Notification Settings
    'xep_0494',  # Client Access Management
    'xep_0502',  # MUC Activity Indicator
    'xep_0511',  # Link Metadata
    'xep_0513',  # Explicit Mentions
    # Meant to be imported by plugins
]

__all__ = PLUGINS + [
    'PluginManager',
    'PluginNotFound',
    'BasePlugin',
    'register_plugin',
    'load_plugin',
]




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Erik Reuterborg Larsson, Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.stanza import Message
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0184 import stanza, Request, Received


class XEP_0184(BasePlugin):

    """
    XEP-0184: Message Delivery Receipts
    """

    name = 'xep_0184'
    description = 'XEP-0184: Message Delivery Receipts'
    dependencies = {'xep_0030'}
    stanza = stanza
    default_config = {
        'auto_ack': True,
        'auto_request': False
    }

    ack_types = ('normal', 'chat', 'headline')

    def plugin_init(self):
        register_stanza_plugin(Message, Request)
        register_stanza_plugin(Message, Received)

        self.xmpp.add_filter('out', self._filter_add_receipt_request)

        self.xmpp.register_handler(
                Callback('Message Receipt',
                    StanzaPath('message/receipt'),
                    self._handle_receipt_received))

        self.xmpp.register_handler(
                Callback('Message Receipt Request',
                    StanzaPath('message/request_receipt'),
                    self._handle_receipt_request))

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature('urn:xmpp:receipts')
        self.xmpp.del_filter('out', self._filter_add_receipt_request)
        self.xmpp.remove_handler('Message Receipt')
        self.xmpp.remove_handler('Message Receipt Request')

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('urn:xmpp:receipts')

    def ack(self, msg):
        """
        Acknowledge a message by sending a receipt.

        Arguments:
            msg -- The message to acknowledge.
        """
        ack = self.xmpp.Message()
        ack['to'] = msg['from']
        ack['receipt'] = msg['id']
        ack.send()

    def _handle_receipt_received(self, msg):
        self.xmpp.event('receipt_received', msg)

    def _handle_receipt_request(self, msg):
        """
        Auto-ack message receipt requests if ``self.auto_ack`` is ``True``.

        Arguments:
            msg -- The incoming message requesting a receipt.
        """
        if self.auto_ack:
            if msg['type'] in self.ack_types:
                if not msg['receipt']:
                    self.ack(msg)

    def _filter_add_receipt_request(self, stanza):
        """
        Auto add receipt requests to outgoing messages, if:

            - ``self.auto_request`` is set to ``True``
            - The message is not for groupchat
            - The message does not contain a receipt acknowledgment
            - The recipient is a bare JID or, if a full JID, one
              that has the ``urn:xmpp:receipts`` feature enabled

        The disco cache is checked if a full JID is specified in
        the outgoing message, which may mean a round-trip disco#info
        delay for the first message sent to the JID if entity caps
        are not used.
        """

        if not self.auto_request:
            return stanza

        if not isinstance(stanza, Message):
            return stanza

        if stanza['request_receipt']:
            return stanza

        if stanza['type'] not in self.ack_types:
            return stanza

        if stanza['receipt']:
            return stanza

        if not stanza['body']:
            return stanza

        if stanza['to'].resource:
            if not self.xmpp['xep_0030'].supports(stanza['to'],
                    feature='urn:xmpp:receipts',
                    cached=True):
                return stanza

        stanza['request_receipt'] = True
        return stanza




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp import JID
from slixmpp.xmlstream import ElementBase, register_stanza_plugin


class Bookmarks(ElementBase):
    name = 'storage'
    namespace = 'storage:bookmarks'
    plugin_attrib = 'bookmarks'
    interfaces = set()

    def add_conference(self, jid, nick, name=None, autojoin=None, password=None):
        conf = Conference()
        conf['jid'] = jid
        conf['nick'] = nick
        if name is None:
            name = jid
        conf['name'] = name
        conf['autojoin'] = autojoin
        conf['password'] = password
        self.append(conf)

    def add_url(self, url, name=None):
        saved_url = URL()
        saved_url['url'] = url
        if name is None:
            name = url
        saved_url['name'] = name
        self.append(saved_url)


class Conference(ElementBase):
    name = 'conference'
    namespace = 'storage:bookmarks'
    plugin_attrib = 'conference'
    plugin_multi_attrib = 'conferences'
    interfaces = {'nick', 'password', 'autojoin', 'jid', 'name'}
    sub_interfaces = {'nick', 'password'}

    def get_autojoin(self):
        value = self._get_attr('autojoin')
        return value in ('1', 'true')

    def set_autojoin(self, value):
        del self['autojoin']
        if value in ('1', 'true', True):
            self._set_attr('autojoin', 'true')

    def set_jid(self, value):
        del self['jid']
        if isinstance(value, JID):
            value = value.full
        self._set_attr('jid', value)


class URL(ElementBase):
    name = 'url'
    namespace = 'storage:bookmarks'
    plugin_attrib = 'url'
    plugin_multi_attrib = 'urls'
    interfaces = {'url', 'name'}


register_stanza_plugin(Bookmarks, Conference, iterable=True)
register_stanza_plugin(Bookmarks, URL, iterable=True)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission
from slixmpp.stanza import Iq
from slixmpp.xmlstream import (
    ElementBase,
    register_stanza_plugin,
)
from slixmpp.plugins.xep_0421.stanza import OccupantId
from slixmpp.plugins.xep_0424.stanza import Retract, Retracted


NS = 'urn:xmpp:message-moderate:1'


class Moderate(ElementBase):
    """
    Moderate request element.

    .. code-block:: xml

          <moderate id="stanza-id-1" xmlns='urn:xmpp:message-moderate:1'>
            <retract xmlns='urn:xmpp:message-retract:1'/>
            <reason>This message contains inappropriate content for this forum</reason>
          </moderate>

    """
    namespace = NS
    name = 'moderate'
    plugin_attrib = 'moderate'
    interfaces = {'id', 'reason'}
    sub_interfaces = {'reason'}


class Moderated(ElementBase):
    """
    Moderated message notification element.

    .. code-block:: xml

        <moderated by='room@muc.example.com/macbeth' xmlns='urn:xmpp:message-moderate:1'>
          <occupant-id xmlns="urn:xmpp:occupant-id:0" id="dd72603deec90a38ba552f7c68cbcc61bca202cd" />
        </moderated>

    """
    namespace = NS
    name = 'moderated'
    plugin_attrib = 'moderated'
    interfaces = {'by'}


def register_plugins():
    # for moderation requests
    register_stanza_plugin(Iq, Moderate)
    register_stanza_plugin(Moderate, Retract)

    # for moderation events
    register_stanza_plugin(Retract, Moderated)
    register_stanza_plugin(Moderated, OccupantId)

    # for tombstones
    register_stanza_plugin(Retracted, Moderated)



import logging

from slixmpp.plugins import BasePlugin

from . import stanza

log = logging.getLogger(__name__)


class XEP_0234(BasePlugin):

    """
    XEP-0234: Jingle File Transfer

    Minimum needed for xep 0385 (Stateless inline media sharing)
    """

    name = "xep_0234"
    description = "XEP-0234: Jingle File Transfer"
    dependencies = {"xep_0082", "xep_0300"}
    stanza = stanza



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Erik Reuterborg Larsson
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0080.stanza import Geoloc
from slixmpp.plugins.xep_0080.geoloc import XEP_0080


register_plugin(XEP_0080)

__all__ = ['Geoloc', 'XEP_0080']



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin
from slixmpp.plugins.xep_0369.stanza import (
    MIX,  Setnick, Join, Leave, Subscribe, Unsubscribe, UpdateSubscription,
    Create, Participant, Destroy,
)
from slixmpp.plugins.xep_0369.mix_core import XEP_0369

register_plugin(XEP_0369)

__all__ = ['MIX', 'Setnick', 'Join', 'Leave', 'Subscribe', 'Unsubscribe',
           'UpdateSubscription', 'Create', 'Participant',
           'Destroy', 'XEP_0369']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from datetime import datetime, timezone

from slixmpp import JID
from slixmpp.stanza import Presence
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.xep_0319 import stanza


def get_local_timezone():
    return datetime.now(timezone.utc).astimezone().tzinfo


class XEP_0319(BasePlugin):
    name = 'xep_0319'
    description = 'XEP-0319: Last User Interaction in Presence'
    dependencies = {'xep_0012'}
    stanza = stanza

    def plugin_init(self):
        self._idle_stamps = {}
        register_stanza_plugin(Presence, stanza.Idle)
        self.api.register(self._set_idle, 'set_idle', default=True)
        self.api.register(self._get_idle, 'get_idle', default=True)
        self.xmpp.register_handler(Callback(
            'Idle Presence',
            StanzaPath('presence/idle'),
            self._idle_presence
        ))
        self.xmpp.add_filter('out', self._stamp_idle_presence)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('urn:xmpp:idle:1')

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature='urn:xmpp:idle:1')
        self.xmpp.del_filter('out', self._stamp_idle_presence)
        self.xmpp.remove_handler('Idle Presence')

    async def idle(self, jid: JID | None = None,
                   since: datetime | None = None):
        """Set an idle duration for a JID

        .. versionchanged:: 1.8.0
            This function is now a coroutine.
        """
        seconds = None
        timezone = get_local_timezone()
        if since is None:
            since = datetime.now(timezone)
        else:
            seconds = datetime.now(timezone) - since
        await self.api['set_idle'](jid, None, None, since)
        await self.xmpp['xep_0012'].set_last_activity(jid=jid, seconds=seconds)

    async def active(self, jid: JID | None = None):
        """Reset the idle timer.

        .. versionchanged:: 1.8.0
            This function is now a coroutine.
        """
        await self.api['set_idle'](jid, None, None, None)
        await self.xmpp['xep_0012'].del_last_activity(jid)

    def _set_idle(self, jid, node, ifrom, data):
        self._idle_stamps[jid] = data

    def _get_idle(self, jid, node, ifrom, data):
        return self._idle_stamps.get(jid, None)

    def _idle_presence(self, pres):
        self.xmpp.event('presence_idle', pres)

    async def _stamp_idle_presence(self, stanza):
        if isinstance(stanza, Presence):
            since = await self.api['get_idle'](stanza['from'] or self.xmpp.boundjid)
            if since:
                stanza['idle']['since'] = since
        return stanza



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp import JID, Iq
from slixmpp.plugins import BasePlugin
from slixmpp.stanza.roster import RosterItem
from slixmpp.plugins.xep_0405 import stanza
from slixmpp.plugins.xep_0369 import stanza as mix_stanza


BASE_NODES = [
    'urn:xmpp:mix:nodes:messages',
    'urn:xmpp:mix:nodes:participants',
    'urn:xmpp:mix:nodes:info',
]


class XEP_0405(BasePlugin):
    '''XEP-0405: MIX-PAM'''

    name = 'xep_0405'
    description = 'XEP-0405: MIX-PAM'
    dependencies = {'xep_0369'}
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self) -> None:
        stanza.register_plugins()

    async def check_server_capability(self) -> bool:
        """Check if the server is MIX-PAM capable"""
        result = await self.xmpp.plugin['xep_0030'].get_info(jid=self.xmpp.boundjid.bare)
        features = result['disco_info']['features']
        return stanza.NS in features

    async def join_channel(self, room: JID, nick: str, subscribe: set[str] | None = None, *,
                           ito: JID | None = None,
                           ifrom: JID | None = None,
                           **iqkwargs) -> set[str]:
        """
        Join a MIX channel.

        :param JID room: JID of the MIX channel
        :param str nick: Desired nickname on that channel
        :param set[str] subscribe: Set of nodes to subscribe to when joining.
            If empty, all nodes will be subscribed by default.

        :rtype: set[str]
        :return: The nodes that failed to subscribe, if any
        """
        if subscribe is None:
            subscribe = set(BASE_NODES)
        if ito is None:
            ito = self.xmpp.boundjid.bare
        iq = self.xmpp.make_iq_set(ito=ito, ifrom=ifrom)
        iq['client_join']['channel'] = room
        iq['client_join']['mix_join']['nick'] = nick
        for node in subscribe:
            sub = mix_stanza.Subscribe()
            sub['node'] = node
            iq['client_join']['mix_join'].append(sub)
        result = await iq.send(**iqkwargs)
        result_nodes = {sub['node'] for sub in result['client_join']['mix_join']}
        return subscribe.difference(result_nodes)

    async def leave_channel(self, room: JID, *,
                            ito: JID | None = None,
                            ifrom: JID | None = None,
                            **iqkwargs) -> Iq:
        """"
        Leave a MIX channel
        :param JID room: JID of the channel to leave
        """
        if ito is None:
            ito = self.xmpp.boundjid.bare
        iq = self.xmpp.make_iq_set(ito=ito, ifrom=ifrom)
        iq['client_leave']['channel'] = room
        iq['client_leave'].enable('mix_leave')
        return await iq.send(**iqkwargs)

    async def get_mix_roster(self, *,
                            ito: JID | None = None,
                            ifrom: JID | None = None,
                            **iqkwargs) -> tuple[list[RosterItem], list[RosterItem]]:
        """
        Get the annotated roster, with MIX channels.

        :return: A tuple of (contacts, mix channels) as RosterItem elements
        """
        iq = self.xmpp.make_iq_get(ito=ito, ifrom=ifrom)
        iq['roster'].enable('annotate')
        result = await iq.send(**iqkwargs)
        self.xmpp.event("roster_update", result)
        contacts = []
        mix = []
        for item in result['roster']:
            channel = item.get_plugin('channel', check=True)
            if channel:
                mix.append(item)
            else:
                contacts.append(item)
        return (contacts, mix)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.stanza import Presence
from slixmpp.types import RosterState, ResourceDict, RosterDBProtocol, JidStr

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from slixmpp import BaseXMPP


class RosterItem:

    """
    A RosterItem is a single entry in a roster node, and tracks
    the subscription state and user annotations of a single JID.

    Roster items may use an external datastore to persist roster data
    across sessions. Client applications will not need to use this
    functionality, but is intended for components that do not have their
    roster persisted automatically by the XMPP server.

    Roster items provide many methods for handling incoming presence
    stanzas that ensure that response stanzas are sent according to
    RFC 3921.

    The external datastore is accessed through a provided interface
    object which is stored in self.db. The interface object MUST
    provide two methods: load and save, both of which are responsible
    for working with a single roster item. A private dictionary,
    self._db_state, is used to store any metadata needed by the
    interface, such as the row ID of a roster item, etc.

    Interface for self.db.load:
        load(owner_jid, jid, db_state):
          owner_jid  -- The JID that owns the roster.
          jid        -- The JID of the roster item.
          db_state   -- A dictionary containing any data saved
                        by the interface object after a save()
                        call. Will typically have the equivalent
                        of a 'row_id' value.

    Interface for self.db.save:
        save(owner_jid, jid, item_state, db_state):
          owner_jid  -- The JID that owns the roster.
          jid        -- The JID of the roster item.
          item_state -- A dictionary containing the fields:
                        'from', 'to', 'pending_in', 'pending_out',
                        'whitelisted', 'subscription', 'name',
                        and 'groups'.
          db_state   -- A dictionary provided for persisting
                        datastore specific information. Typically,
                        a value equivalent to 'row_id' will be
                        stored here.

    State Fields:
        from         -- Indicates if a subscription of type 'from'
                        has been authorized.
        to           -- Indicates if a subscription of type 'to' has
                        been authorized.
        pending_in   -- Indicates if a subscription request has been
                        received from this JID and it has not been
                        authorized yet.
        pending_out  -- Indicates if a subscription request has been sent
                        to this JID and it has not been accepted yet.
        subscription -- Returns one of: 'to', 'from', 'both', or 'none'
                        based on the states of from, to, pending_in,
                        and pending_out. Assignment to this value does
                        not affect the states of the other values.
        whitelisted  -- Indicates if a subscription request from this
                        JID should be automatically accepted.
        name         -- A user supplied alias for the JID.
        groups       -- A list of group names for the JID.

    Attributes:
        xmpp        -- The main Slixmpp instance.
        owner       -- The JID that owns the roster.
        jid         -- The JID for the roster item.
        db          -- Optional datastore interface object.
        last_status -- The last presence sent to this JID.
        resources   -- A dictionary of online resources for this JID.
                       Will contain the fields 'show', 'status',
                       and 'priority'.

    Methods:
        load                -- Retrieve the roster item from an
                               external datastore, if one was provided.
        save                -- Save the roster item to an external
                               datastore, if one was provided.
        remove              -- Remove a subscription to the JID and revoke
                               its whitelisted status.
        subscribe           -- Subscribe to the JID.
        authorize           -- Accept a subscription from the JID.
        unauthorize         -- Deny a subscription from the JID.
        unsubscribe         -- Unsubscribe from the JID.
        send_presence       -- Send a directed presence to the JID.
        send_last_presence  -- Resend the last sent presence.
        handle_available    -- Update the JID's resource information.
        handle_unavailable  -- Update the JID's resource information.
        handle_subscribe    -- Handle a subscription request.
        handle_subscribed   -- Handle a notice that a subscription request
                               was authorized by the JID.
        handle_unsubscribe  -- Handle an unsubscribe request.
        handle_unsubscribed -- Handle a notice that a subscription was
                               removed by the JID.
        handle_probe        -- Handle a presence probe query.
    """

    xmpp: 'BaseXMPP'
    jid: JidStr
    owner: JidStr
    db: RosterDBProtocol | None
    _state: RosterState
    resources: dict[str, ResourceDict]
    _db_state: dict[str, Any]
    last_status: Presence | None

    def __init__(self, xmpp: 'BaseXMPP', jid: JidStr, owner: JidStr | None = None,
                 state: RosterState | None = None, db: RosterDBProtocol | None = None, roster=None):
        """
        Create a new roster item.

        :param xmpp: The main Slixmpp instance.
        :param jid: The item's JID.
        :param owner: The roster owner's JID. Defaults
                      so self.xmpp.boundjid.bare.
        :param state: A dictionary of initial state values.
        :param db: An optional interface to an external datastore.
        :param roster: The roster object containing this entry.
        """
        self.xmpp = xmpp
        self.jid = jid
        self.owner = owner or self.xmpp.boundjid.bare
        self.last_status = None
        self.resources = {}
        self.roster = roster
        self.db = db
        if state:
            self._state = state
        else:
            self._state = RosterState({
                'from': False,
                'to': False,
                'pending_in': False,
                'pending_out': False,
                'whitelisted': False,
                'subscription': 'none',
                'name': '',
                'groups': [],
                'removed': False,
            })

        self._db_state = {}
        self.load()

    def set_backend(self, db: RosterDBProtocol | None=None, save: bool = True) -> None:
        """
        Set the datastore interface object for the roster item.

        :param db: The new datastore interface.
        :param save: If True, save the existing state to the new
                     backend datastore. Defaults to True.
        """
        self.db = db
        if save:
            self.save()
        self.load()

    def load(self) -> RosterState | None:
        """
        Load the item's state information from an external datastore,
        if one has been provided.
        """
        if self.db:
            item = self.db.load(self.owner, self.jid,
                                       self._db_state)
            if item:
                self['name'] = item['name']
                self['groups'] = item['groups']
                self['from'] = item['from']
                self['to'] = item['to']
                self['whitelisted'] = item['whitelisted']
                self['pending_out'] = item['pending_out']
                self['pending_in'] = item['pending_in']
                self['subscription'] = self._subscription()
            return self._state
        return None

    def save(self, remove: bool = False) -> None:
        """
        Save the item's state information to an external datastore,
        if one has been provided.

        :param remove: If True, expunge the item from the datastore.
        """
        self['subscription'] = self._subscription()
        if remove:
            self._state['removed'] = True
        if self.db:
            self.db.save(self.owner, self.jid,
                         self._state, self._db_state)

        # Finally, remove the in-memory copy if needed.
        if remove:
            del self.xmpp.roster[self.owner][self.jid]

    def __getitem__(self, key):
        """Return a state field's value."""
        if key in self._state:
            if key == 'subscription':
                return self._subscription()
            return self._state[key]
        else:
            raise KeyError

    def __setitem__(self, key: str, value: str | list[str] | bool) -> None:
        """
        Set the value of a state field.

        For boolean states, the values True, 'true', '1', 'on',
        and 'yes' are accepted as True; all others are False.

        Arguments:
            key   -- The state field to modify.
            value -- The new value of the state field.
        """
        if key in self._state:
            if key in ['name', 'subscription', 'groups']:
                self._state[key] = value  # type: ignore
            else:
                value = str(value).lower()
                self._state[key] = value in ('true', '1', 'on', 'yes')  # type: ignore
        else:
            raise KeyError

    def _subscription(self):
        """Return the proper subscription type based on current state."""
        if self['to'] and self['from']:
            return 'both'
        elif self['from']:
            return 'from'
        elif self['to']:
            return 'to'
        else:
            return 'none'

    def remove(self) -> None:
        """
        Remove a JID's whitelisted status and unsubscribe if a
        subscription exists.
        """
        if self['to']:
            p = self.xmpp.Presence()
            p['to'] = self.jid
            p['type'] = 'unsubscribe'
            if self.xmpp.is_component:
                p['from'] = self.owner
            p.send()
            self['to'] = False
        self['whitelisted'] = False
        self.save()

    def subscribe(self) -> None:
        """Send a subscription request to the JID."""
        p = self.xmpp.Presence()
        p['to'] = self.jid
        p['type'] = 'subscribe'
        if self.xmpp.is_component:
            p['from'] = self.owner
        self['pending_out'] = True
        self.save()
        p.send()

    def authorize(self) -> None:
        """Authorize a received subscription request from the JID."""
        self['from'] = True
        self['pending_in'] = False
        self.save()
        self._subscribed()
        self.send_last_presence()

    def unauthorize(self) -> None:
        """Deny a received subscription request from the JID."""
        self['from'] = False
        self['pending_in'] = False
        self.save()
        self._unsubscribed()
        p = self.xmpp.Presence()
        p['to'] = self.jid
        p['type'] = 'unavailable'
        if self.xmpp.is_component:
            p['from'] = self.owner
        p.send()

    def _subscribed(self) -> None:
        """Handle acknowledging a subscription."""
        p = self.xmpp.Presence()
        p['to'] = self.jid
        p['type'] = 'subscribed'
        if self.xmpp.is_component:
            p['from'] = self.owner
        p.send()

    def unsubscribe(self) -> None:
        """Unsubscribe from the JID."""
        p = self.xmpp.Presence()
        p['to'] = self.jid
        p['type'] = 'unsubscribe'
        if self.xmpp.is_component:
            p['from'] = self.owner
        self.save()
        p.send()

    def _unsubscribed(self) -> None:
        """Handle acknowledging an unsubscribe request."""
        p = self.xmpp.Presence()
        p['to'] = self.jid
        p['type'] = 'unsubscribed'
        if self.xmpp.is_component:
            p['from'] = self.owner
        p.send()

    def send_presence(self, **kwargs):
        """
        Create, initialize, and send a Presence stanza.

        If no recipient is specified, send the presence immediately.
        Otherwise, forward the send request to the recipient's roster
        entry for processing.

        :param pshow: The presence's show value.
        :param pstatus: The presence's status message.
        :param ppriority: This connections' priority.
        :param pto: The recipient of a directed presence.
        :param pfrom: The sender of a directed presence, which should
                      be the owner JID plus resource.
        :param ptype: The type of presence, such as 'subscribe'.
        :param pnick: Optional nickname of the presence's sender.
        """
        if self.xmpp.is_component and not kwargs.get('pfrom', ''):
            kwargs['pfrom'] = self.owner
        if not kwargs.get('pto', ''):
            kwargs['pto'] = self.jid
        self.xmpp.send_presence(**kwargs)

    def send_last_presence(self) -> None:
        if self.last_status is None:
            pres = self.roster.last_status
            if pres is None:
                self.send_presence()
            else:
                pres['to'] = self.jid
                if self.xmpp.is_component:
                    pres['from'] = self.owner
                else:
                    del pres['from']
                pres.send()
        else:
            self.last_status.send()

    def handle_available(self, presence: Presence) -> None:
        resource = presence['from'].resource
        data = ResourceDict({
            'status': presence['status'],
            'show': presence['show'],
            'priority': presence['priority'],
        })
        got_online = not self.resources
        if resource not in self.resources:
            self.resources[resource] = {}
        old_status = self.resources[resource].get('status', '')
        old_show = self.resources[resource].get('show', None)
        self.resources[resource].update(data)
        if got_online:
            self.xmpp.event('got_online', presence)
        if old_show != presence['show'] or old_status != presence['status']:
            self.xmpp.event('changed_status', presence)

    def handle_unavailable(self, presence: Presence) -> None:
        resource = presence['from'].resource
        if not self.resources:
            return
        if resource in self.resources:
            del self.resources[resource]
        self.xmpp.event('changed_status', presence)
        if not self.resources:
            self.xmpp.event('got_offline', presence)

    def handle_subscribe(self, presence: Presence) -> None:
        """
        +------------------------------------------------------------------+
        |  EXISTING STATE          |  DELIVER?  |  NEW STATE               |
        +------------------------------------------------------------------+
        |  "None"                  |  yes       |  "None + Pending In"     |
        |  "None + Pending Out"    |  yes       |  "None + Pending Out/In" |
        |  "None + Pending In"     |  no        |  no state change         |
        |  "None + Pending Out/In" |  no        |  no state change         |
        |  "To"                    |  yes       |  "To + Pending In"       |
        |  "To + Pending In"       |  no        |  no state change         |
        |  "From"                  |  no *      |  no state change         |
        |  "From + Pending Out"    |  no *      |  no state change         |
        |  "Both"                  |  no *      |  no state change         |
        +------------------------------------------------------------------+
        """
        if self.xmpp.is_component:
            if not self['from'] and not self['pending_in']:
                self['pending_in'] = True
                self.xmpp.event('roster_subscription_request', presence)
            elif self['from']:
                self._subscribed()
            self.save()
        else:
            #server shouldn't send an invalid subscription request
            if not self['from'] and not self['pending_in']:
                self['pending_in'] = True
            self.xmpp.event('roster_subscription_request', presence)

    def handle_subscribed(self, presence: Presence) -> None:
        """
        +------------------------------------------------------------------+
        |  EXISTING STATE          |  DELIVER?  |  NEW STATE               |
        +------------------------------------------------------------------+
        |  "None"                  |  no        |  no state change         |
        |  "None + Pending Out"    |  yes       |  "To"                    |
        |  "None + Pending In"     |  no        |  no state change         |
        |  "None + Pending Out/In" |  yes       |  "To + Pending In"       |
        |  "To"                    |  no        |  no state change         |
        |  "To + Pending In"       |  no        |  no state change         |
        |  "From"                  |  no        |  no state change         |
        |  "From + Pending Out"    |  yes       |  "Both"                  |
        |  "Both"                  |  no        |  no state change         |
        +------------------------------------------------------------------+
        """
        if self.xmpp.is_component:
            if not self['to'] and self['pending_out']:
                self['pending_out'] = False
                self['to'] = True
                self.xmpp.event('roster_subscription_authorized', presence)
            self.save()
        else:
            self.xmpp.event('roster_subscription_authorized', presence)

    def handle_unsubscribe(self, presence: Presence) -> None:
        """
        +------------------------------------------------------------------+
        |  EXISTING STATE          |  DELIVER?  |  NEW STATE               |
        +------------------------------------------------------------------+
        |  "None"                  |  no        |  no state change         |
        |  "None + Pending Out"    |  no        |  no state change         |
        |  "None + Pending In"     |  yes *     |  "None"                  |
        |  "None + Pending Out/In" |  yes *     |  "None + Pending Out"    |
        |  "To"                    |  no        |  no state change         |
        |  "To + Pending In"       |  yes *     |  "To"                    |
        |  "From"                  |  yes *     |  "None"                  |
        |  "From + Pending Out"    |  yes *     |  "None + Pending Out     |
        |  "Both"                  |  yes *     |  "To"                    |
        +------------------------------------------------------------------+
        """
        if self.xmpp.is_component:
            if not self['from'] and self['pending_in']:
                self['pending_in'] = False
                self._unsubscribed()
            elif self['from']:
                self['from'] = False
                self._unsubscribed()
                self.xmpp.event('roster_subscription_remove', presence)
            self.save()
        else:
            if not self['from'] and self['pending_in']:
                self['pending_in'] = False
            elif self['from']:
                self['from'] = False
            self.xmpp.event('roster_subscription_remove', presence)

    def handle_unsubscribed(self, presence: Presence) -> None:
        """
        +------------------------------------------------------------------+
        |  EXISTING STATE          |  DELIVER?  |  NEW STATE               |
        +------------------------------------------------------------------+
        |  "None"                  |  no        |  no state change         |
        |  "None + Pending Out"    |  yes       |  "None"                  |
        |  "None + Pending In"     |  no        |  no state change         |
        |  "None + Pending Out/In" |  yes       |  "None + Pending In"     |
        |  "To"                    |  yes       |  "None"                  |
        |  "To + Pending In"       |  yes       |  "None + Pending In"     |
        |  "From"                  |  no        |  no state change         |
        |  "From + Pending Out"    |  yes       |  "From"                  |
        |  "Both"                  |  yes       |  "From"                  |
        +------------------------------------------------------------------
        """
        if self.xmpp.is_component:
            if not self['to'] and self['pending_out']:
                self['pending_out'] = False
            elif self['to'] and not self['pending_out']:
                self['to'] = False
                self.xmpp.event('roster_subscription_removed', presence)
            self.save()
        else:
            self.xmpp.event('roster_subscription_removed', presence)

    def handle_probe(self, presence: Presence) -> None:
        if self['from']:
            self.send_last_presence()
        if self['pending_out']:
            self.subscribe()
        if not self['from']:
            self._unsubscribed()

    def reset(self) -> None:
        """
        Forgot current resource presence information as part of
        a roster reset request.
        """
        self.resources = {}

    def __repr__(self) -> str:
        return repr(self._state)




# slixmpp.xmlstream.handler.waiter
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Part of Slixmpp: The Slick XMPP Library
# :copyright: (c) 2011 Nathanael C. Fritz
# :license: MIT, see LICENSE for more details
from __future__ import annotations

import logging
from asyncio import Event, wait_for, TimeoutError
from typing import TYPE_CHECKING

import slixmpp
from slixmpp.xmlstream.stanzabase import StanzaBase
from slixmpp.xmlstream.handler.base import BaseHandler
from slixmpp.xmlstream.matcher.base import MatcherBase

if TYPE_CHECKING:
    from slixmpp.xmlstream.xmlstream import XMLStream

log = logging.getLogger(__name__)


class Waiter(BaseHandler):

    """
    The Waiter handler allows an event handler to block until a
    particular stanza has been received. The handler will either be
    given the matched stanza, or ``False`` if the waiter has timed out.

    :param string name: The name of the handler.
    :param matcher: A :class:`~slixmpp.xmlstream.matcher.base.MatcherBase`
                    derived object for matching stanza objects.
    :param stream: The :class:`~slixmpp.xmlstream.xmlstream.XMLStream`
                   instance this handler should monitor.
    """
    _event: Event

    def __init__(self, name: str, matcher: MatcherBase, stream: XMLStream | None = None):
        BaseHandler.__init__(self, name, matcher, stream=stream)
        self._event = Event()

    def prerun(self, payload: StanzaBase) -> None:
        """Store the matched stanza when received during processing.

        :param payload: The matched
            :class:`~slixmpp.xmlstream.stanzabase.StanzaBase` object.
        """
        if not self._event.is_set():
            self._event.set()
            self._payload = payload

    def run(self, payload: StanzaBase) -> None:
        """Do not process this handler during the main event loop."""
        pass

    async def wait(self, timeout: int | None = None) -> StanzaBase | None:
        """Block an event handler while waiting for a stanza to arrive.

        Be aware that this will impact performance if called from a
        non-threaded event handler.

        Will return either the received stanza, or ``False`` if the
        waiter timed out.

        :param int timeout: The number of seconds to wait for the stanza
            to arrive. Defaults to the the stream's
            :class:`~slixmpp.xmlstream.xmlstream.XMLStream.response_timeout`
            value.
        """
        stream_ref = self.stream
        if stream_ref is None:
            raise ValueError('wait() called without a stream')
        stream = stream_ref()
        if stream is None:
            raise ValueError('wait() called without a stream')
        if timeout is None:
            timeout = slixmpp.xmlstream.RESPONSE_TIMEOUT

        try:
            await wait_for(
                self._event.wait(), timeout,
            )
        except TimeoutError:
            log.warning("Timed out waiting for %s", self.name)
        stream.remove_handler(self.name)
        return self._payload

    def check_delete(self) -> bool:
        """Always remove waiters after use."""
        return True



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp import JID
from slixmpp.plugins import BasePlugin
from slixmpp.exceptions import IqError, IqTimeout

MUC_ROOMINFO = 'http://jabber.org/protocol/muc#roominfo'


class XEP_0502(BasePlugin):
    """
    XEP-0502: MUC Activity Indicator
    """

    name = "xep_0502"
    description = "XEP-0502: MUC Activity Indicator"
    dependencies = {"xep_0030", "xep_0128"}
    namespace = 'urn:xmpp:muc-activity'

    async def get_activity(self, jid: JID, **iqargs) -> float | None:
        """
        Return the activity of a room, or None if the activity is not found
        """
        try:
            info_iq = await self.xmpp.plugin['xep_0030'].get_info(
                jid=jid,
                **iqargs,
            )
        except (IqError, IqTimeout):
            return None
        disco = info_iq.get_plugin('disco_info', check=True)
        if not disco:
            return None
        if 'forms' not in disco:
            return None
        forms = disco['forms']
        if not forms:
            return None
        field = '{%s}message-activity' % self.namespace
        for form in forms:
            values = form.get_values()
            if values.get('FORM_TYPE') == [MUC_ROOMINFO]:
                return values.get(field, None)
        return None




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.


from slixmpp import Message
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0235 import stanza, OAuth


class XEP_0235(BasePlugin):

    name = 'xep_0235'
    description = 'XEP-0235: OAuth Over XMPP'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, OAuth)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('urn:xmpp:oauth:0')

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature='urn:xmpp:oauth:0')



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Erik Reuterborg Larsson
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0059.stanza import Set
from slixmpp.plugins.xep_0059.rsm import ResultIterator, XEP_0059


register_plugin(XEP_0059)

__all__ = ['Set', 'ResultIterator', 'XEP_0059']



import logging
from datetime import datetime
from pathlib import Path
from typing import Iterable, Literal

from slixmpp.plugins import BasePlugin
from slixmpp.stanza import Message
from slixmpp.xmlstream import register_stanza_plugin

from . import stanza

log = logging.getLogger(__name__)


class XEP_0447(BasePlugin):

    """
    XEP-0447: Stateless File Sharing

    Only support outgoing SFS, incoming is not handled at all.
    """

    name = "xep_0447"
    description = "XEP-0447: Stateless File Sharing"
    dependencies = {"xep_0300", "xep_0446"}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, stanza.StatelessFileSharing)

        register_stanza_plugin(stanza.StatelessFileSharing, stanza.Sources)
        register_stanza_plugin(
            stanza.StatelessFileSharing, self.xmpp["xep_0446"].stanza.File
        )
        register_stanza_plugin(stanza.Sources, stanza.UrlData, iterable=True)

    def get_sfs(
        self,
        path: Path,
        uris: Iterable[str] | None = None,
        media_type: str | None = None,
        desc: str | None = None,
        disposition: Literal["inline", "attachment"] | None = None
    ) -> stanza.StatelessFileSharing:
        """
        Produce an SFS element from a file present locally.

        :param path: Path of the file.
        :param uris: Iterable on uris to that file.
        :param media_type: Media type of the file.
        :param desc: Description of the file.
        :param disposition: The content-disposition of the file.
        :returns: The SFS element.
        """
        sfs = stanza.StatelessFileSharing()
        if disposition:
            sfs["disposition"] = disposition
        if uris is not None:
            for uri in uris:
                ref = stanza.UrlData()
                ref["target"] = uri
                sfs["sources"].append(ref)
        if media_type:
            sfs["file"]["media-type"] = media_type
        if desc:
            sfs["file"]["desc"] = desc
        sfs["file"]["name"] = path.name

        stat = path.stat()
        sfs["file"]["size"] = stat.st_size
        sfs["file"]["date"] = datetime.fromtimestamp(stat.st_mtime)

        h = self.xmpp.plugin["xep_0300"].compute_hash(path)
        h["value"] = h["value"].decode()
        sfs["file"].append(h)

        return sfs




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class IPCheck(ElementBase):

    name = 'ip'
    namespace = 'urn:xmpp:sic:0'
    plugin_attrib = 'ip_check'
    interfaces = {'ip_check'}
    is_extension = True

    def get_ip_check(self):
        return self.xml.text

    def set_ip_check(self, value):
        if value:
            self.xml.text = value
        else:
            self.xml.text = ''

    def del_ip_check(self):
        self.xml.text = ''




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.xmlstream import ElementBase


class Register(ElementBase):

    namespace = 'jabber:iq:register'
    name = 'query'
    plugin_attrib = 'register'
    interfaces = {'username', 'password', 'email', 'nick', 'name',
                  'first', 'last', 'address', 'city', 'state', 'zip',
                  'phone', 'url', 'date', 'misc', 'text', 'key',
                  'registered', 'remove', 'instructions', 'fields'}
    sub_interfaces = interfaces
    form_fields = {'username', 'password', 'email', 'nick', 'name',
                   'first', 'last', 'address', 'city', 'state', 'zip',
                   'phone', 'url', 'date', 'misc', 'text', 'key'}

    def get_registered(self):
        present = self.xml.find('{%s}registered' % self.namespace)
        return present is not None

    def get_remove(self):
        present = self.xml.find('{%s}remove' % self.namespace)
        return present is not None

    def set_registered(self, value):
        if value:
            self.add_field('registered')
        else:
            del self['registered']

    def set_remove(self, value):
        if value:
            self.add_field('remove')
        else:
            del self['remove']

    def add_field(self, value):
        self._set_sub_text(value, '', keep=True)

    def get_fields(self):
        fields = set()
        for field in self.form_fields:
            if self.xml.find('{%s}%s' % (self.namespace, field)) is not None:
                fields.add(field)
        return fields

    def set_fields(self, fields):
        del self['fields']
        for field in fields:
            self._set_sub_text(field, '', keep=True)

    def del_fields(self):
        for field in self.form_fields:
            self._del_sub(field)


class RegisterFeature(ElementBase):

    name = 'register'
    namespace = 'http://jabber.org/features/iq-register'
    plugin_attrib = name
    interfaces = set()
